/*
	Copyright (C) 2005 by Ruben Zilibowitz <rzilibowitz@users.sourceforge.net>
	Part of the Toy Cars Project http://toycars.sourceforge.net

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License.
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY.

	See the COPYING file for more details.
*/

/*
 *  Created by Ruben Zilibowitz on 1/01/05.
 */

#include "CarGame.h"
#include "GameConstants.h"
#include "Rectangle.h"
#include "ScException.h"
#include "Kaboom.h"
#include "Startline.h"
#include "glf.h"
#include "TcPreferences.h"
#include <cstdlib>

extern TcPreferences *gPreferences; // the maximum number of particles in the animation at any one time
const int kNumLaps = 4;         // the number of laps to the race

// patch for drand48 and seed48 on platforms that dont support these function calls
#ifndef seed48
unsigned short * seed48(unsigned short xseed[3])
{
	srandom(*((unsigned long*)(xseed+1)));
	return NULL;
}
#endif

#ifndef drand48
const double kInvRandomRange = 1.0 / 2147483647.0;
double drand48()
{
	return random() * kInvRandomRange;
}
#endif

extern TcPreferences *gPreferences;

class PathLine : public ScPanel
{
public:
	PathLine(CarGame &inGame)
		: ScPanel(inGame.getTilemap().getMapWidth() * inGame.getTilemap().getTileWidth(), inGame.getTilemap().getMapHeight() * inGame.getTilemap().getTileHeight()),
		  pathPolygon(inGame.pathPolygon)
	{}
	~PathLine() {}
	
protected:
	const list<Tuple> *pathPolygon;
	
	void draw()
	{
		glColor3f(1.0, 0.0, 0.0);
		glDisable(GL_TEXTURE_2D);
		glEnable(GL_TEXTURE_1D);
		glBegin(GL_LINE_LOOP);
		for (list<Tuple>::const_iterator i = pathPolygon->begin(); i != pathPolygon->end(); i++) {
			glVertex2d(i->x * kScale, i->y * kScale);
		}
		glEnd();
	}
};

// -----------------------------------------------------------------------------
//	CarGame::CarGame
//  
//  Constructor. Sets up animation. Inits random number generator.
// -----------------------------------------------------------------------------

CarGame::CarGame(ScTilemap* inTilemap, Player** inPlayers, int inNumPlayers, Tuple inStartLineA, Tuple inStartLineB, const list<Tuple> *inPathPolygon)
	:   tilemap(*inTilemap), players(inPlayers), numPlayers(inNumPlayers),
		particles(new ScParticles(gPreferences->getMaxParticles(), 0.351, 0.091, 0.031)),
		runningTime(0.0), startLineA(inStartLineA), startLineB(inStartLineB), pathPolygon(inPathPolygon)
{
	int i;
	
	// set the game rect
	gameRect.left = 0;
	gameRect.bottom = 0;
	gameRect.right = tilemap.getMapWidth() * tilemap.getTileWidth();
	gameRect.top = tilemap.getMapHeight() * tilemap.getTileHeight();
	
	// Set up Scoobie animation
	    
	// put the tile map inside the animation
	animation.push_back(inTilemap);
	// put the particles inside the animation
	animation.push_back(particles);
	// create and add start line
	animation.push_back(new Startline(startLineA, startLineB));
	
	// put the cars inside the top layer
	// and put players into playersInGame list
	for (i = 0; i < numPlayers; i++)
	{
		animation.push_back(&players[i]->car);
		playersInGame.push_back(players[i]);
	}
	
	// load needle image
	needle = new NeedleImage();
	
	// the glf library should have already been initialised
    
	// create and load the minimap
	minimap = new Minimap(tilemap, playersInGame);
	
	// path line
	//animation.push_back(new PathLine(*this));
	
	// init seed for random number generator
	unsigned short s[4];
	time((time_t*)s);
	*(clock_t*)(s+2) = clock();
	seed48(s + 1);
	
	setupStart();
}

// -----------------------------------------------------------------------------
//	CarGame::setupStart
//  
//  Sets up start locations of cars
// -----------------------------------------------------------------------------

void CarGame::setupStart()
{
	// compute intersection of the path with the start line
	Tuple A = startLineA / kScale;
	Tuple B = startLineB / kScale;
	Tuple E, F;
	Tuple result(0,0);
	double s, t, divisor;
	list<Tuple>::const_iterator i, j;
	for (i = pathPolygon->begin(), (j = pathPolygon->end())--; i != pathPolygon->end(); j = i++) {
		E = *i;
		F = *j;
		divisor = A.x*(F.y-E.y)+B.x*(E.y-F.y)+(B.y-A.y)*F.x+(A.y-B.y)*E.x;
		s = -(B.x*(F.y-E.y)+E.x*(B.y-F.y)+(E.y-B.y)*F.x) / divisor;	// along path AB
		t = (A.x*(F.y-B.y)+B.x*(A.y-F.y)+(B.y-A.y)*F.x) / divisor;	// along path EF
		if (0 <= s && s < 1 && 0 <= t && t < 1) {
			result = s*A + (1-s)*B;
			break;
		}
	}
	
	// set start locations of all cars
	// we setup a two by two line behind the start line perpendicular to it
	Tuple vOffset = (A - B).Direction().perp();
	Tuple hOffset = vOffset.perp();
	double theta = atan2(-vOffset.y, -vOffset.x);
	Tuple placement;
	for (int k = 0; k < numPlayers; k++)
	{
		if (k % 2 == 0) {
			result += vOffset * players[k]->car.getWidth() / kScale;
			placement = result - hOffset * players[k]->car.getHeight() / kScale * 0.55;
		}
		else {
			placement = result + hOffset * players[k]->car.getHeight() / kScale * 0.55;
		}
		players[k]->car.setLocation(placement);
		players[k]->car.setOrientation(theta);
		players[k]->computePathLocation();
	}
}

// -----------------------------------------------------------------------------
//	CarGame::~CarGame
//  
//  Destructor. Cleans up.
// -----------------------------------------------------------------------------

CarGame::~CarGame()
{
	// delete players
	for (int i = 0; i < numPlayers; i++)
		delete players[i];
	delete [] players;
    
	// delete the views
	for (list<ViewFollow>::iterator i = views.begin(); i != views.end(); i++)
		delete i->view;
    
	// delete the minimap
	delete minimap;
	
	// delete the needle image
	delete needle;
	
	// delete the path polygon
	delete pathPolygon;
}

// -----------------------------------------------------------------------------
//	CarGame::addView
//  
//  Adds a view to the game with a given player to follow.
// -----------------------------------------------------------------------------

void CarGame::addView(ScPortRect& portRect, int playerFollow)
{
	ScRect viewRect = {0, 0, portRect.width, portRect.height};
	ScView* viewP = new ScView(portRect, viewRect, animation);
	views.push_back(ViewFollow(viewP, playerFollow));
}

// -----------------------------------------------------------------------------
//	CarGame::doEvent
//  Receive an event for a particular player.
// -----------------------------------------------------------------------------

void CarGame::doEvent(PlayerEvent* event, unsigned short playerNum)
{
    if (playerNum < numPlayers)
			players[playerNum]->doEvent(event);
}

// -----------------------------------------------------------------------------
//	CarGame::step
//  
//  Calculate and draw a single frame of the game being played.
// -----------------------------------------------------------------------------

void CarGame::step(double dt)
{
	computePhysics(dt);
	animation.process(dt);
	
	for (list<Player*>::iterator i = playersInGame.begin(); i != playersInGame.end(); i++)
		(*i)->updateLaps(startLineA, startLineB, runningTime);
	
	runningTime += dt;
	
	drawViews();
    drawTime(runningTime);
	minimap->draw(120, 20, 1.0);
}

// -----------------------------------------------------------------------------
//	CarGame::renderEveryTexture
//
//  This should be called once before the beginning of the game. This function
//  renders all textures in the game once, which is necessary to load them into
//  VRAM. Without calling this, the game may run slowly at the beginning until
//  it has finally drawn every texture.
// -----------------------------------------------------------------------------

void CarGame::renderEveryTexture()
{
	int i;
	ScRect& viewRect = views.front().view->getViewRect();
	WheelSprite* wheelSprite = players[0]->car.getLeftWheelSprite();
    ToyCar* carSprite;
    
    glLoadIdentity();
	glOrtho(0, viewRect.right - viewRect.left, 0, viewRect.top - viewRect.bottom, -1.0, 1.0);
	
	// render all the car sprite frames
    for (i = 0; i < numPlayers; i++)
    {
        carSprite = &players[i]->car;
        carSprite->moveTo(0,0);
        for (i = 0; i < carSprite->getNumFrames(); i++)
        {
            carSprite->setFrame(i);
            carSprite->doDraw();
        }
	}
    
	// render all the wheel sprite frames
	wheelSprite->moveTo(0,0);
	for (i = 0; i < wheelSprite->getNumFrames(); i++)
	{
		wheelSprite->setFrame(i);
		wheelSprite->doDraw();
	}
		
	// render the needle image
	needle->setDegrees(0);
	needle->draw();
	
	// render the minimap image
	minimap->draw(0, 0, 1.0);
}

// -----------------------------------------------------------------------------
//	CarGame::computePhysics
//  
//  Performs integration on forces for each car. Then checks for collisions
//  between cars and wall tiles. Then checks for collisions between all cars
//  in game. Finally calls resolveCollisions.
// -----------------------------------------------------------------------------

const double kStepFrequency = 500.0;	// number of integrations per second

void CarGame::computePhysics(double dt)
{
	list<Player*>::iterator playerI, playerJ;
	
	// perform numerical integration on vehicles
	// generate dust particles and integrate user inputs
	int j, steps = int(ceil(dt * kStepFrequency));
	double t, small_dt = dt / steps;
	for (playerI = playersInGame.begin(); playerI != playersInGame.end(); playerI++)
	{
		// integrate car dynamics and keyboard input and generate dust particles
		t = 0.0;
		for (j = 0; j < steps; j++)
		{
			(*playerI)->integrateKeys(small_dt);
			(*playerI)->car.rk2_step(small_dt);
			
			t += small_dt;
			generateDustFromWheel((*playerI)->car.getWheel_FL(), small_dt, t);
			generateDustFromWheel((*playerI)->car.getWheel_FR(), small_dt, t);
			generateDustFromWheel((*playerI)->car.getWheel_BL(), small_dt, t);
			generateDustFromWheel((*playerI)->car.getWheel_BR(), small_dt, t);
		}
		
		// update geometry
		(*playerI)->car.compute_geom_world_coords();
	}
	
	// check for collisions
	list<Contact> contacts;
	for (playerI = playersInGame.begin(); playerI != playersInGame.end(); playerI++)
	{
		findTileContacts((*playerI)->car, contacts);
		for ((playerJ = playerI)++; playerJ != playersInGame.end(); playerJ++)
		{
			if (Geom::collide((*playerI)->car.getGeom_wc(), (*playerJ)->car.getGeom_wc(), contacts))
			{
                collisionAction(contacts, *playerI, *playerJ);
			}
		}
	}
	resolveCollisions(contacts);
}

// -----------------------------------------------------------------------------
//	CarGame::drawViews
//  
//  Scrolls views. Bounds the scrolling to the game rect. Invokes the draw
//  function for each view. Draws additional things for each view such as
//  speedometre and engine rpm of the corresponding player.
// -----------------------------------------------------------------------------

void CarGame::drawViews()
{
	for (list<ViewFollow>::iterator i = views.begin(); i != views.end(); i++)
	{
		// scroll the view to the car location
		
		i->view->scrollTo(	floor(players[i->follow]->car.getLocation_wc().x * kScale),
							floor(players[i->follow]->car.getLocation_wc().y * kScale));
		
		// bound the view to the animation bounds
		
		if (i->view->getViewRect().left < gameRect.left)
			i->view->scrollBy(gameRect.left - i->view->getViewRect().left, 0);
		
		if (i->view->getViewRect().bottom < gameRect.bottom)
			i->view->scrollBy(0, gameRect.bottom - i->view->getViewRect().bottom);
		
		if (i->view->getViewRect().right > gameRect.right)
			i->view->scrollBy(gameRect.right - i->view->getViewRect().right, 0);
		
		if (i->view->getViewRect().top > gameRect.top)
			i->view->scrollBy(0, gameRect.top - i->view->getViewRect().top);
		
		// draw scene
		i->view->draw();
		
		//
		// draw display
		//
		glLoadIdentity();
		glOrtho(0, i->view->getViewRect().right - i->view->getViewRect().left, 0, i->view->getViewRect().top - i->view->getViewRect().bottom, -1, 1);
		drawDisplay(players[i->follow]);
	}
}

void CarGame::drawDisplay(Player *player)
{
	drawSpeedometer(player);
	drawRPMMeter(player);
	drawLapCount(player);
	if (player->getLaps() >= kNumLaps)
		drawFinishTime(player);
}

void CarGame::drawSpeedometer(Player *player)
{
	glPushMatrix();
	needle->setDegrees(player->car.getVelocity().x*5 - 90.0);
	glTranslatef(100, 64, 0);
	needle->draw();
	glPopMatrix();
}

void CarGame::drawRPMMeter(Player *player)
{
	glPushMatrix();
	needle->setDegrees(player->car.getRPM()*0.1 - 90.0);
	glTranslatef(200, 64, 0);
	needle->draw();
	glPopMatrix();
}

void CarGame::drawLapCount(Player *player)
{
	static char str15[16];
	glDisable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);
	glColor4f(1.0f,1.0f,1.0f,0.5f);
	glPushMatrix();
	glTranslatef(24, gPreferences->getScreenHeight()-48, 0);
	glScalef(10, 10, 1);
	sprintf(str15, "Laps: %d", player->getLaps());
	glfDrawSolidString(str15);
	glPopMatrix();
}

void CarGame::drawFinishTime(Player *player)
{
	static char str15[16];
	glPushMatrix();
	glScalef(12, 12, 1);
	glTranslatef(2, gPreferences->getScreenHeight()/24.0 + 1, 0);
	sprintf(str15, "Finish time: %.2lf s", player->getCourseTime());
	glfDrawSolidString(str15);
	glTranslatef(0, -2, 0);
	sprintf(str15, "Best lap time: %.2lf s", player->getBestLapTime());
	glfDrawSolidString(str15);
	glPopMatrix();
}

// -----------------------------------------------------------------------------
//	CarGame::drawTime
//  
//  Takes time in seconds and renders clock digits
// -----------------------------------------------------------------------------

void CarGame::drawTime(double theTime)
{
	static char str31[32];
	
	glDisable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);
	glColor4f(1.0f,1.0f,1.0f,0.5f);
	
	glLoadIdentity();
	glOrtho(0, gPreferences->getScreenWidth(), 0, gPreferences->getScreenHeight(), -1, 1);
	glViewport(0, 0, gPreferences->getScreenWidth(), gPreferences->getScreenHeight());
	glTranslatef(24, gPreferences->getScreenHeight()-24, 0);
	
	int minutes = int(theTime) / 60;
	sprintf(str31, "%d:%.2lf", minutes, theTime - 60*minutes);
	glScalef(12, 12, 1);
	glfDrawSolidString(str31);
}

// -----------------------------------------------------------------------------
//	CarGame::findTileContacts
//
//  Compute contacts for any collisions between car and tiles.
// -----------------------------------------------------------------------------

void CarGame::findTileContacts(ToyCar& car, list<Contact>& contacts)
{
    int firstrow = int(car.getBounds().bottom) / tilemap.getTileHeight();
    int lastrow = int(car.getBounds().top) / tilemap.getTileHeight();
    int firstcol = int(car.getBounds().left) / tilemap.getTileWidth();
    int lastcol = int(car.getBounds().right) / tilemap.getTileWidth();
    int row, col;
	
	firstrow = std::max(firstrow, 0);
	lastrow = std::min(lastrow, (tilemap.getMapHeight() - 1));
	firstcol = std::max(firstcol, 0);
	lastcol = std::min(lastcol, (tilemap.getMapWidth() - 1));
	
    list<block> blocks;
    
    for (row = firstrow; row <= lastrow; row++)
    {
        for (col = firstcol; col <= lastcol; col++)
        {
            if (tilemap(col,row) == 25)
            {
                blocks.push_back(block(col+1,col,row+1,row));
            }
        }
    }
    
    sewSeams(blocks);

    for (list<block>::const_iterator i = blocks.begin(); i != blocks.end(); i++)
    {
        Rectangle block(	i->right * tilemap.getTileWidth() / kScale,
							i->left * tilemap.getTileWidth() / kScale,
							i->top * tilemap.getTileHeight() / kScale,
							i->bottom * tilemap.getTileHeight() / kScale);
		
        Geom::collide(car.getGeom_wc(), block, contacts);
	}
}

// -----------------------------------------------------------------------------
//	CarGame::sewSeams
//
//  This algorithm reduces a list of blocks to one where there are no 'seams'.
//  A seam is two blocks sharing a common edge. Applying this to the list
//  significantly improves collision detection with the blocks.
//
//  If we do not apply this algorithm to the list of blocks, then crashing
//  into a straight wall sometimes behaves as if the car is crashing into the
//  corner of one of the tiles in the wall, producing a very undesirable effect.
// -----------------------------------------------------------------------------

void CarGame::sewSeams(list<block>& blocks)
{
    bool seam_found;
    list<block>::iterator tileA, tileB;
    
    do
    {
        seam_found = false;
        
        for (tileA = blocks.begin(); tileA != blocks.end() && !seam_found; tileA++)
        {
            for ((tileB = tileA)++; tileB != blocks.end() && !seam_found; tileB++)
            {
                // check for seams between tileA and tileB
                
                if (tileA->top == tileB->top && tileA->bottom == tileB->bottom)
                {
                    if (tileA->right == tileB->left)
                    {
                        tileB->left = tileA->left;
                        blocks.erase(tileA);
                        seam_found = true;
                    }
                    else if (tileB->right == tileA->left)
                    {
                        tileB->right = tileA->right;
                        blocks.erase(tileA);
                        seam_found = true;
                    }
                }
                else if (tileA->left == tileB->left && tileA->right == tileB->right)
                {
                    if (tileA->top == tileB->bottom)
                    {
                        tileB->bottom = tileA->bottom;
                        blocks.erase(tileA);
                        seam_found = true;
                    }
                    else if (tileB->top == tileA->bottom)
                    {
                        tileB->top = tileA->top;
                        blocks.erase(tileA);
                        seam_found = true;
                    }
                }
            }
        }
    }
    while (seam_found);
}

// -----------------------------------------------------------------------------
//	CarGame::resolveCollisions
//  
//  Attempts to resolve all the collisions in the given contact list.
//  An iterative method is used. Overlap error is compensated for by
//  translating bodies away from each other in resolve_overlap().
// -----------------------------------------------------------------------------

void CarGame::resolveCollisions(list<Contact>& contacts)
{
    list<Contact>::iterator c;
	bool collisions;
	unsigned char maxIterations = 20;
	
	// resolve collisions
	
    do
    {
		collisions = false;
        for (c = contacts.begin(); c != contacts.end(); c++)
        {
            if (c->resolve_collision())
				collisions = true;
        }
    }
	while (collisions && --maxIterations);
    
	// seperate overlaps
	
    for (c = contacts.begin(); c != contacts.end(); c++)
    {
        c->resolve_overlap();
    }
}

// -----------------------------------------------------------------------------
//	CarGame::generateDustFromWheel
//  
//  Computes the dust particles sprayed out by the given wheel over a given
//  time period dt (usually very short).
//  
//  Time of creation of the particles can be moved backwards by specifying a
//  positive value for rewind. The reason for this is because this function may
//  be called at many sub-frame times for each frame (ie during integration of
//  forces and velocities). One wants to synchronise the particles however, as
//  if they were created on the actual start of the frame. Otherwise, they will
//  look not right.
// -----------------------------------------------------------------------------

const double kDustProliferation = 25.0;
const double kDustLife = 1.2;

void CarGame::generateDustFromWheel(const Wheel& wheel, double dt, double rewind)
{
	int numParticles, i;
	Tuple lat, longt, loc;
	const Vehicle& vehicle = wheel.getVehicle();
	double dustVel;
	
	// local class rround
	// rounds off x based on probability from fractional part
	class _rround
	{
	public:
		int operator() (double x)
		{
			if (x - floor(x) > drand48())
				return (int)(x) + 1;
			else
				return (int)(x);
		}
	}rround;
	
	// tyre location in world coords
	loc = (vehicle.getLocation_wc() + vehicle.getNorm().fromLocal(wheel.getLocation_vc())) * kScale;
	
	// dust direction in world coords
	lat = vehicle.getNorm().fromLocal(wheel.getNorm());
	longt = -lat.perp();
	
	// dust from longtitudinal slip
	dustVel = wheel.getLongtitudinalSlipVelocity();
	if (fabs(dustVel) > 5.0)
	{
		numParticles = rround(dt * 2.0*kDustProliferation * fabs(dustVel));
		//if (numParticles) particles->newParticle(Particle(loc.x, loc.y, 10.0, 0, 0, 0));
		for (i = 0; i < numParticles; i++)
			particles->newParticle(Particle(loc.x, loc.y, kDustLife, longt.x + (drand48()-0.5)/dustVel*4.0, longt.y + (drand48()-0.5)/dustVel*4.0, dustVel * kScale).step(-rewind));
	}
	
	// dust from lateral slip
	dustVel = wheel.getVelocity().y;
	if (fabs(dustVel) > 5.0)
	{
		numParticles = rround(dt * 3.0*kDustProliferation * fabs(dustVel));
		//if (numParticles) particles->newParticle(Particle(loc.x, loc.y, 10.0, 0, 0, 0));
		for (i = 0; i < numParticles; i++)
			particles->newParticle(Particle(loc.x, loc.y, kDustLife, lat.x + (drand48()-0.5)/dustVel*4.0, lat.y + (drand48()-0.5)/dustVel*4.0, dustVel * kScale).step(-rewind));
	}
}
