/*
	Copyright (C) 2005 by Ruben Zilibowitz <rzilibowitz@users.sourceforge.net>
	Part of the Toy Cars Project http://toycars.sourceforge.net

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License.
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY.

	See the COPYING file for more details.
*/

/*
 *  Created by Ruben on Wed Apr 21 2004.
 */

#include "Contact.h"
#include "RigidBody.h"

const double kRestitution = 0.5;
const double kThreshold = -0.1;
const double kSeperation = 0.15;

RigidBody gNullbody;

//------------------------------------------------------------------------------------------------------
// impulse computation
// ------------------------------------------
//
//									-(1+e)*(rel_norm_v)
//			j =  ------------------------------------------------------------
//			       (1/Ma + 1/Mb) + [(ra x norm)^2/Ia] + [(rb x norm)^2/Ib]
//------------------------------------------------------------------------------------------------------

Contact::Contact(RigidBody* inBodyA, RigidBody* inBodyB, const Tuple& inPa, const Tuple& inPb, const Tuple& inNormal)
    :   bodyA(inBodyA ? inBodyA : &gNullbody), bodyB(inBodyB ? inBodyB : &gNullbody),
        pa(inPa), pb(inPb), normal(inNormal),
        ra(pa - bodyA->mLocation_wc),
        rb(pb - bodyB->mLocation_wc),
        K(bodyA->mInvMass + bodyB->mInvMass + sqr(ra^normal)*bodyA->mInvInertia + sqr(rb^normal)*bodyB->mInvInertia)
{
}

bool Contact::resolve_collision()
{
    double epsilon;
    double rel_norm_v = compute_velocity();
    
    //
    // determine type of collision
    //
    
    if (rel_norm_v < kThreshold)
        epsilon = kRestitution; // normal collision
    else if (rel_norm_v < 0.0)
        epsilon = 1.0;          // microcollision
    else
        return false;                 // parting contact (no collision)
        
    //
    // apply impulse
    //
    
    double j = -(1 + epsilon) * rel_norm_v / K;
    Tuple impulse = j*normal;
    
    bodyA->mVelocity_wc += bodyA->mInvMass * impulse;
    bodyA->mAngularVelocity += bodyA->mInvInertia * (ra^impulse);
    
    bodyB->mVelocity_wc -= bodyB->mInvMass * impulse;
    bodyB->mAngularVelocity -= bodyB->mInvInertia * (rb^impulse);
	
	return true;
}

void Contact::resolve_overlap()
{
    double overlap = (pa - pb)*normal;
    
    //
    // check for overlap and resolve
    //
    
    if (overlap < 0)
    {
        double s = -kSeperation * overlap / K;
        Tuple shift = s*normal;
        
        bodyA->mLocation_wc += bodyA->mInvMass * shift;
        bodyA->mOrientation_wc += bodyA->mInvInertia * (ra^shift);
        
        bodyB->mLocation_wc -= bodyB->mInvMass * shift;
        bodyB->mOrientation_wc -= bodyB->mInvInertia * (rb^shift);
    }
}

double Contact::compute_velocity()
{
    Tuple va = bodyA->mVelocity_wc + (bodyA->mAngularVelocity * ra.perp());
    Tuple vb = bodyB->mVelocity_wc + (bodyB->mAngularVelocity * rb.perp());
    
    return (va - vb)*normal;
}
