/*
	Copyright (C) 2005 by Ruben Zilibowitz <rzilibowitz@users.sourceforge.net>
	Part of the Toy Cars Project http://toycars.sourceforge.net

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License.
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY.

	See the COPYING file for more details.
*/

/*
 *  Created by Ruben on Wed Aug 25 2004.
 */

#include "Convex.h"

// if sin^2 of an angle between two edges is less than this, they are contacting edges.
const double kEdgeEdgeThreshold = 0.0001;   // 0.0003 is just less than 1 degree

// function declarations
void crop_edges(const Tuple& N, Tuple& P1, Tuple& Q1, Tuple& P2, Tuple& Q2);

Convex::Convex(const Tuple* array, int size)
{
    for (int n = 0; n < size; n++)
        push_back(array[n]);
}


// check for collisions between two geoms.
bool Convex::collide(const Convex& geomA, const Convex& geomB, list<Contact>& contacts)
{
    const Convex* geom_touching = &geomA;
    const Convex* geom_receiving = &geomB;
    
    double depth;
    Convex::const_iterator Va, Vb, P;
    Tuple N;
    
    double bestdepth;
    Convex::const_iterator bestVa, bestVb, bestP;
    Tuple bestN;
    
    // initialise variables
    bestdepth = HUGE_VAL;
    
	for(Va = geomB.begin(), Vb = geomB.last(); Va != geomB.end(); Vb = Va++)
    {
        // compute normal direction. (length is not normalised)
        N = (*Vb - *Va).perp();
        
        // calculate deepest point of other polygon along this normal
		P = geomA.deepest_point(N);
		
        // the penetration depth
        depth = (*P - *Vb)*N;
        
        // the edge separates the polygons, they are therefore disjoint
        if (depth > 0)
            return false;
		
        // convert to the square of the normalised depth
        depth = sqr(depth)/(N*N);
        
        // is it smallest penetration found?
		if (depth < bestdepth)
        {
            bestVa = Va;
            bestVb = Vb;
            bestP = P;
            bestN = N;
            bestdepth = depth;
        }
    }
    
	// check geomB against edges of geomA
	for(Va = geomA.begin(), Vb = geomA.last(); Va != geomA.end(); Vb = Va++)
    {
        // compute normal direction. (length is not normalised)
        N = (*Vb - *Va).perp();
        
        // calculate deepest point of other polygon along this normal
		P = geomB.deepest_point(N);
		
        // the penetration depth
        depth = (*P - *Vb)*N;
        
        // the edge separates the polygons, they are therefore disjoint
        if (depth > 0)
            return false;
		
        // convert to the square of the normalised depth
        depth = sqr(depth)/(N*N);
		
        // is it smallest penetration found?
		if (depth < bestdepth)
        {
            bestVa = Va;
            bestVb = Vb;
            bestP = P;
            bestN = N;
            bestdepth = depth;
            geom_touching = &geomB;
            geom_receiving = &geomA;
        }
    }
    
    // Normalise
    
    bestN.Normalise();
    
    // Determine type of collision (edge/edge or vertex/edge)
    
    Convex::const_iterator nextP = geom_touching->next(bestP);
    Convex::const_iterator prevP = geom_touching->prev(bestP);
    
	Tuple PA = *nextP - *bestP;
    Tuple BP = *bestP - *prevP;
        
    if (sqr(PA*bestN)/(PA*PA) < kEdgeEdgeThreshold)   // Check for edge/edge collision
    {
        Tuple P1 = *nextP;
        Tuple Q1 = *bestP;
        Tuple P2 = *bestVa;
        Tuple Q2 = *bestVb;
        
        crop_edges(bestN, P1, Q1, P2, Q2);
        
        contacts.push_back(Contact(geom_touching->parent, geom_receiving->parent, P1, Q2, bestN));
        contacts.push_back(Contact(geom_touching->parent, geom_receiving->parent, Q1, P2, bestN));
    }
    else if (sqr(BP*bestN)/(BP*BP) < kEdgeEdgeThreshold)  // Check for edge/edge collision
    {
        Tuple P1 = *bestP;
        Tuple Q1 = *prevP;
        Tuple P2 = *bestVa;
        Tuple Q2 = *bestVb;
        
        crop_edges(bestN, P1, Q1, P2, Q2);
        
        contacts.push_back(Contact(geom_touching->parent, geom_receiving->parent, P1, Q2, bestN));
        contacts.push_back(Contact(geom_touching->parent, geom_receiving->parent, Q1, P2, bestN));
    }
    else    // it is a vertex/edge collision
    {
        Tuple edge = *bestVa - *bestVb;
        double lambda = ((*bestP - *bestVb)*edge)/(edge*edge);
        Tuple Q = (*bestVb) + lambda*edge;
        
        contacts.push_back(Contact(geom_touching->parent, geom_receiving->parent, *bestP, Q, bestN));
    }
    
    return true;
}

// find the vertex which lies deepest against the direction given as N.
Convex::const_iterator Convex::deepest_point(const Tuple& N) const
{
    double d;
	double spanmin = HUGE_VAL;
        
    // compute the deepest penetrating point
    	
	const_iterator minV;
	
    for(const_iterator V = begin(); V != end(); V++)
    {
        d = (*V)*N;
		
        if (d < spanmin)
		{
			spanmin = d;
			minV = V;
		}
    }
    
    return minV;
}

// crop_edges takes two line segments given by (P1..Q1) and (P2..Q2) and crops them
// down so that (P1..Q2) is in same direction as N and (Q1..P2) is in same direction as N.
// it is assumed that they are oriented oppositely and are roughly at same angle.
void crop_edges(const Tuple& N, Tuple& P1, Tuple& Q1, Tuple& P2, Tuple& Q2)
{
    double lambda;
    
    Tuple newP1 = P1;
    Tuple newQ2 = Q2;
    
    lambda = (N^(Q2 - Q1))/(N^(P1 - Q1));
    if (lambda < 1)
    {
        newP1 = lambda*(P1 - Q1) + Q1;
    }
    else
    {
        lambda = (N^(P1 - Q2))/(N^(P2 - Q2));
        newQ2 = lambda*(P2 - Q2) + Q2;
    }
    
    lambda = (N^(P2 - Q1))/(N^(P1 - Q1));
    if (lambda > 0)
    {
        Q1 = lambda*(P1 - Q1) + Q1;
    }
    else
    {
        lambda = (N^(Q1 - Q2))/(N^(P2 - Q2));
        P2 = lambda*(P2 - Q2) + Q2;
    }
    
    P1 = newP1;
    Q2 = newQ2;
}
