/*
	Copyright (C) 2005 by Ruben Zilibowitz <rzilibowitz@users.sourceforge.net>
	Part of the Toy Cars Project http://toycars.sourceforge.net

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License.
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY.

	See the COPYING file for more details.
*/

/*
 *  Created by Ruben on Wed Apr 21 2004.
 */
 

#include "RigidBody.h"
#include "Contact.h"

const double kTwoPi = 2 * M_PI;

RigidBody::RigidBody()
    : mMass(HUGE_VAL),
      mInertia(HUGE_VAL),
      
      mInvMass(0),
      mInvInertia(0),
      
      mTotalForce_wc(Tuple(0,0)),
      mTotalTorque(0),
      
      mLocation_wc(Tuple(0,0)),
      mVelocity_wc(Tuple(0,0)),
      mAcceleration_wc(Tuple(0,0)),
      
      mOrientation_wc(0),
      mAngularVelocity(0),
      mAngularAcceleration(0),
      
      mNorm(-sin(0.0), cos(0.0)),
      
      mGeom(Geom()),
      mGeom_wc(Geom())
{
}

RigidBody::RigidBody(   double inMass,
						double inInertia,
						Tuple inLocation_wc,
						Tuple inVelocity_wc,
						double inOrientation_wc,
						double inAngularVelocity,
						const Geom& inGeom)
    : mMass(inMass),
      mInertia(inInertia),
      
      mInvMass(inMass > 1e6 ? 0 : 1/inMass),
      mInvInertia(inInertia > 1e6 ? 0 : 1/inInertia),
      
      mTotalForce_wc(Tuple(0,0)),
      mTotalTorque(0),
      
      mLocation_wc(inLocation_wc),
      mVelocity_wc(inVelocity_wc),
      mAcceleration_wc(Tuple(0,0)),
      
      mOrientation_wc(inOrientation_wc),
      mAngularVelocity(inAngularVelocity),
      mAngularAcceleration(0),
      
      mNorm(-sin(inOrientation_wc), cos(inOrientation_wc)),
      
      mGeom(inGeom),
      mGeom_wc(inGeom)
{
    mGeom_wc.setParent(this);
    compute_geom_world_coords();
    
    while (mOrientation_wc < 0)
        mOrientation_wc += kTwoPi;
    while (mOrientation_wc >= kTwoPi)
        mOrientation_wc -= kTwoPi;
}

void RigidBody::compute_force_and_torque()
{
    // default - simulation of gravity
    // bodies with high mass are considered "anchored" and will not move

    mTotalForce_wc.x = 0;
    
    if (mInvMass > 0)
        mTotalForce_wc.y = -9.8;
    else
        mTotalForce_wc.y = 0;
    
    mTotalTorque = 0;
}

// integrate velocity and acceleration - method used is 2nd order Runge-Kutta
void RigidBody::rk2_step(double dt)
{
    Tuple saveLoc = mLocation_wc;
    double saveOrient = mOrientation_wc;
    Tuple saveVel = mVelocity_wc;
    double saveAngVel = mAngularVelocity;
    
    // use half step
    
        // integrate to get P'(t + 0.5)
    mLocation_wc += 0.5 * dt * mVelocity_wc;
    mOrientation_wc += 0.5 * dt * mAngularVelocity;
	
		// integrate to get P(t + 0.5)
    mVelocity_wc += 0.5 * dt * mAcceleration_wc;
    mAngularVelocity += 0.5 * dt * mAngularAcceleration;
    
        // bound the angle to 0..2pi
    if (mOrientation_wc < 0)
        mOrientation_wc += kTwoPi;
    else if (mOrientation_wc >= kTwoPi)
        mOrientation_wc -= kTwoPi;
    
        // compute sine and cosine
    mNorm.x = -sin(mOrientation_wc);
    mNorm.y = cos(mOrientation_wc);
    
        // compute P''(t + 0.5)
    compute_force_and_torque();
    mAcceleration_wc = mInvMass * mTotalForce_wc;
	mAngularAcceleration = mInvInertia * mTotalTorque;
    
    // use full step with half way derivatives
    
        // integrate to get P'(t + 1)
    mLocation_wc = saveLoc + (dt * mVelocity_wc);
    mOrientation_wc = saveOrient + (dt * mAngularVelocity);
	
		// integrate to get P(t + 1)
    mVelocity_wc = saveVel + (dt * mAcceleration_wc);
    mAngularVelocity = saveAngVel + (dt * mAngularAcceleration);
    
        // bound the angle to 0..2pi
    if (mOrientation_wc < 0)
        mOrientation_wc += kTwoPi;
    else if (mOrientation_wc >= kTwoPi)
        mOrientation_wc -= kTwoPi;
    
        // compute sine and cosine
    mNorm.x = -sin(mOrientation_wc);
    mNorm.y = cos(mOrientation_wc);
    
        // compute P''(t + 1)
    compute_force_and_torque();
    mAcceleration_wc = mInvMass * mTotalForce_wc;
	mAngularAcceleration = mInvInertia * mTotalTorque;
}

void RigidBody::compute_geom_world_coords()
{
    Geom::const_iterator a;
    Geom::iterator b;
    Convex::const_iterator i;
    Convex::iterator j;
    
    for (a = mGeom.begin(), b = mGeom_wc.begin(); a != mGeom.end(); a++, b++)
    {
        for (i = a->begin(), j = b->begin(); i != a->end(); i++, j++)
        {
            *j = mNorm.fromLocal(*i) + mLocation_wc;
        }
    }
}
