/*
	Copyright (C) 2005 by Ruben Zilibowitz <rzilibowitz@users.sourceforge.net>
	Part of the Toy Cars Project http://toycars.sourceforge.net

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License.
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY.

	See the COPYING file for more details.
*/

/*
 *  Created by Ruben on Wed Apr 21 2004.
 */

#ifndef RIGIDBODY_H
#define RIGIDBODY_H

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "Geom.h"

// RigidBody class
// A two-dimensional rigid body with kinematic properties and an associated geometry
// or shape used to check for collisions between two or more different rigid bodies.

class RigidBody
{
friend class Contact;

public:
			RigidBody();
			
			RigidBody(  double inMass,
						double inInertia,
						Tuple inLocation_wc,
						Tuple inVelocity_wc,
						double inOrientation_wc,
						double inAngularVelocity,
						const Geom& inGeom);
    
	virtual ~RigidBody() {}
    
	virtual void compute_force_and_torque();
    
	virtual void rk2_step(double dt);
    
	void compute_geom_world_coords();
    
protected:
	double  mMass;					// kg
	double  mInertia;				// kg m^2
    
    double  mInvMass;               // 1/kg
    double  mInvInertia;            // 1/(kg m^2)
	
	Tuple   mTotalForce_wc;			// N
	double  mTotalTorque;			// Nm
	
	Tuple   mLocation_wc;			// m
	Tuple   mVelocity_wc;			// m/s
	Tuple   mAcceleration_wc;		// m/s/s
	
	double  mOrientation_wc;		// rad
	double  mAngularVelocity;		// rad/s
	double  mAngularAcceleration;   // rad/s/s
	
    Tuple   mNorm;					// normal vector for orientation
    
	const Geom mGeom;       // fixed shape in local coordinates
	Geom mGeom_wc;			// above transformed to world coordinates

public:
	// accessor functions
	
	double	getMass() const { return mMass; }
	double  getInertia() const { return mInertia; }
    
    double  getInvMass() const { return mInvMass; }
    double  getInvInertia() const { return mInvInertia; }
	
	Tuple   getTotalForce_wc() const { return mTotalForce_wc; }
	double  getTotalTorque() const { return mTotalTorque; }
	
	Tuple   getLocation_wc() const { return mLocation_wc; }
	Tuple   getVelocity_wc() const { return mVelocity_wc; }
	Tuple   getAcceleration_wc() const { return mAcceleration_wc; }
	
	double  getOrientation_wc() const { return mOrientation_wc; }
	double  getAngularVelocity() const { return mAngularVelocity; }
	double  getAngularAcceleration() const { return mAngularAcceleration; }
	
    Tuple	getNorm() const { return mNorm; }
    
	const Geom&	getGeom() const { return mGeom; }
	const Geom&	getGeom_wc() const { return mGeom_wc; }
	
	// impulse
	
	void	applyImpulse(Tuple impulse) { mVelocity_wc += impulse; }
    
	void	setLocation(Tuple loc) { mLocation_wc = loc; }
	void	setOrientation(double theta) { mOrientation_wc = theta; }
};

#endif
