/*
	Copyright (C) 2005 by Ruben Zilibowitz <rzilibowitz@users.sourceforge.net>
	Part of the Toy Cars Project http://toycars.sourceforge.net

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License.
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY.

	See the COPYING file for more details.
*/

/*
 *  Created by Ruben on Wed Apr 21 2004.
 *
 *  Some code taken from physics demo by Olivier Renault
 *
 */

#ifndef TUPLE_H
#define TUPLE_H

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <cmath>

template <typename T>
inline T sqr(T val)
{
    return val*val;
}

//
// class Tuple
// A standard multi-purpose vector with two elements. Standard vector operations may
// be performed on the class.

class Tuple
{
public:
	double x, y;

	inline Tuple()
    {}

	inline Tuple(double Ix, double Iy)
        : x(Ix) , y(Iy)
	{}

	inline Tuple &operator /=(const double Scalar)	{ x /= Scalar; y /= Scalar;		return *this; }

	inline Tuple &operator *=(const double Scalar)	{ x *= Scalar; y *= Scalar;		return *this; }
	
	inline Tuple &operator +=(const Tuple &Other) { x += Other.x;	y += Other.y;	return *this; }

	inline Tuple &operator -=(const Tuple &Other) { x -= Other.x;	y -= Other.y;	return *this;	}

	inline double operator ^ (const Tuple &V)	const	{	return (x * V.y) - (y * V.x); } // cross product (perp-dot product)

	inline double operator * (const Tuple &V)	const	{	return (x*V.x) + (y*V.y); } // dot product

	inline Tuple operator * (double  s)			const	{	return Tuple(x*s, y*s); }

	friend inline Tuple operator * (double k, const Tuple& V)  {   return Tuple(V.x*k, V.y*k); }
	
	inline Tuple operator / (double  s)			const	{	return Tuple(x/s, y/s); }
	
	inline Tuple operator + (const Tuple &V)	const	{	return Tuple(x+V.x, y+V.y); }
		
	inline Tuple operator - (const Tuple &V)	const	{	return Tuple(x-V.x, y-V.y); }

	inline Tuple operator -(void) const { return Tuple(-x, -y); }
	
    inline Tuple perp() const { return Tuple(-y, x); }
	
	inline Tuple flip() const { return Tuple(x, -y); }
	inline Tuple flop() const { return Tuple(-x, y); }
	
	inline double Norm() const { return x*x + y*y; }
	inline double Length() const { return sqrt(x*x + y*y); }
    
    inline Tuple toLocal (const Tuple &V) const { return Tuple(y*V.x - x*V.y, x*V.x + y*V.y); }
    
    inline Tuple fromLocal (const Tuple &V) const { return Tuple(y*V.x + x*V.y, -x*V.x + y*V.y); }
    
	double Normalise(void) 
	{	
		double r = Length();	
		
		if (r == 0.0) 
			return 0.0; 
		
		(*this) *= (1.0 / r); 
	
		return r;	
	}

	Tuple Direction(void) const
	{
		Tuple Temp(*this);

		Temp.Normalise();

		return Temp;
	}

	double Angle(const Tuple& xE) const
	{
		double dot = (*this) * xE;
		double cross = (*this) ^ xE;
		
		// angle between segments
		double angle = atan2(cross, dot);

		return angle;
	}

	Tuple& Rotate(double angle)
	{
		double tx = x;
		x =  x * cos(angle) - y * sin(angle);
		y = tx * sin(angle) + y * cos(angle);

		return *this;
	}

	Tuple& Rotate(const Tuple& xCentre, double fAngle)
	{
		Tuple D = *this - xCentre;
		D.Rotate(fAngle);

		*this = xCentre + D;

		return *this;
	}

	void Clamp(const Tuple& min, const Tuple& max)
	{
		x = (x < min.x)? min.x : (x > max.x)? max.x : x;
		y = (y < min.y)? min.y : (y > max.y)? max.y : y;
	}
};

//
// some useful functions
//

inline int iround(double real)
{
    return (int)floor(real + 0.5);
}

inline double sign(double x)
{
	return (x < 0.0) ? -1.0 : 1.0;
}

inline Tuple unit_vector(double arg)
{
    return Tuple(cos(arg), sin(arg));
}

#endif
