/*
	Copyright (C) 2005 by Ruben Zilibowitz <rzilibowitz@users.sourceforge.net>
	Part of the Toy Cars Project http://toycars.sourceforge.net

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License.
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY.

	See the COPYING file for more details.
*/

/*
 *  Created by Ruben Zilibowitz on 2/01/05.
 */

#include "Player.h"

// constants

const double kShortToFrac = 1.0 / 65535.0;

const double kSteerSpeed = kMaxSteerAngle*1.6;		// rad / s
const double kSteerReturnSpeed = kSteerSpeed*10;	// rad / s

const double kThrottleSpeed = 2.0;						// units / s
const double kThrottleReturnSpeed = kThrottleSpeed*10;	// units / s

const double kBrakeSpeed = 2.0;						// units / s
const double kBrakeReturnSpeed = kBrakeSpeed*10;	// units / s

// member functions

Player::Player(ToyCar& inCar, list<Tuple> *path, int inPlayerNumber)
	: car(inCar), lapFlag(0), laps(-1), curLap(-1e-1), prevLap(-1e-1), revolutions(-1), bestLapTime(INFINITY), courseTime(0.0),
	  pathEnd(path->end()), pathClosestPerp(INFINITY), pathSize(path->size()), playerNumber(inPlayerNumber)
{
	// init key states
	for (int i = 0; i < kNumKeys; i++)
		keys[i] = false;
	
	// compute pathA, pathB, pathIndex here...
	computePathLocation();
}

void Player::computePathLocation()
{
	list<Tuple>::const_iterator Va, Vb;
	int n;
	
	pathClosestPerp = INFINITY;
	for ((Va = pathEnd)++, (Vb = pathEnd)--, n = 0; Va != pathEnd; Vb = Va++, n++)
	{
		if (testPathPoint(Vb))
			pathIndex = n;
		if (testPathSegment(Va, Vb))
			pathIndex = n;
	}
}

Tuple Player::getPathVector()
{
	if (pathA == pathEnd)
	{
		list<Tuple>::const_iterator tempA;
		(tempA = pathB)++;
		if (tempA == pathEnd)
			tempA++;
		return *tempA - *pathB;
	}
	
	return *pathA - *pathB;
}

bool Player::testPathPoint(list<Tuple>::const_iterator Vb)
{
	Tuple loc = car.getLocation_wc();
	Tuple perp = loc - *Vb;
	
	if (perp*perp < pathClosestPerp) {
		pathClosestPerp = perp*perp;
		pathClosestPoint = *Vb;
		pathSegmentDistance = 0;
		pathA = pathEnd;
		pathB = Vb;
		return true;
	}
	return false;
}

bool Player::testPathSegment(list<Tuple>::const_iterator Va, list<Tuple>::const_iterator Vb)
{
	Tuple line = *Va - *Vb;
	Tuple loc = car.getLocation_wc();
	double dist = ((loc - *Vb)*line) / (line*line);
	Tuple perp = loc - (dist*line + (*Vb));
	
	if (dist >= 0 && dist < 1 && perp*perp < pathClosestPerp) {
		pathClosestPerp = perp*perp;
		pathClosestPoint = dist*line + (*Vb);
		pathSegmentDistance = dist;
		pathA = Va;
		pathB = Vb;
		return true;
	}
	return false;
}

void Player::updatePathLocation()
{
	list<Tuple>::const_iterator Va, Vb;
	const list<Tuple>::const_iterator curA = pathA;
	const list<Tuple>::const_iterator curB = pathB;
	const int curIndex = pathIndex;
	
	pathClosestPerp = INFINITY;
	
	if (curA == pathEnd)
	{
		testPathPoint(curB);
		
		(Va = curB)++;
		Vb = curB;
		if (Va == pathEnd)	Va++;
 		testPathSegment(Va, Vb);
		
		Va = curB;
		(Vb = curB)--;
		if (Vb == pathEnd)	Vb--;
		if (testPathSegment(Va, Vb))
			pathIndex = curIndex - 1;
	}
	else
	{
		Va = curA;
		Vb = curB;
		testPathSegment(Va, Vb);
		testPathPoint(Vb);
		if (testPathPoint(Va) == true)
			pathIndex = curIndex + 1;
		
		(Va = curA)++;
		(Vb = curB)++;
		if (Va == pathEnd)	Va++;
		if (Vb == pathEnd)	Vb++;
		
		if (testPathSegment(Va, Vb))
			pathIndex = curIndex + 1;
		
		(Va = curA)--;
		(Vb = curB)--;
		if (Va == pathEnd)	Va--;
		if (Vb == pathEnd)	Vb--;
		
		if (testPathSegment(Va, Vb))
			pathIndex = curIndex - 1;
	}
}

void Player::doEvent(PlayerEvent* event)
{
	//assert(event != NULL);
	
	switch (event->type)
	{
		case kPlayerKeyUp:
			keyUp(((PlayerEventKey*)event)->key);
			break;
		
		case kPlayerKeyDown:
			keyDown(((PlayerEventKey*)event)->key);
			break;
		
		case kPlayerPaddleMove:
			paddleMove(((PlayerEventPaddle*)event)->control, ((PlayerEventPaddle*)event)->value);
			break;
	}
	
	//printf("player %x received an event\n", this);
}

void Player::paddleMove(PlayerControl which, unsigned short value)
{
	switch (which)
	{
		case kSteerWheel:
			car.mSteerWheelAngle = (value*kShortToFrac - 0.5) * kMaxSteerAngle;
			break;
			
		case kThrottle:
			car.mThrottle = (value*kShortToFrac - 0.5) * 2.0;
			break;
			
		case kBrake:
			car.mBrake = value * kShortToFrac;
			break;
		
		default:
			fprintf(stderr, "Player::paddleMove received unknown control type %d\n", which);
	}
}

/* Updates the lap counter for this player. Returns 1 or 0 depending on whether
   the call to this function resulted in the lap count being increased or not.
*/
char Player::updateLaps(const Tuple& lineA, const Tuple& lineB, double curTime)
{
	const Tuple vec1 = car.getLocation_wc()*kScale - lineA;
	const Tuple vec2 = lineB - lineA;
	
	prevLap = curLap;	
	curLap = vec2^vec1;
	
	if (vec1*vec2 < 0)
		return 0;
	
	// check if we crossed over the branch cut in the atan function
	if (curLap > 0 && prevLap <= 0) {
        lapFlag++;
        revolutions++;
		//printf("revolutions changed to %d\n", revolutions);
    }
    else if (curLap < 0 && prevLap >= 0) {
    	lapFlag--;
        revolutions--;
		//printf("revolutions changed to %d\n", revolutions);
    }
	
	if (lapFlag == 1)
	{
		laps++;			// player has done a lap
		lapFlag = 0;
		
		// update time parameters
		if (laps > 0)
			bestLapTime = std::min((curTime - courseTime), bestLapTime);
		courseTime = curTime;
		//printf("courseTime %lf, bestLapTime %lf\n", courseTime, bestLapTime);
		// test
		//printf("course time: %f\nbest lap time: %f\n", courseTime, bestLapTime);
        
        return 1;
	}
	else if (lapFlag == -2)
	{
		lapFlag = -1;	// player has done a lap backwards
	}
    
    return 0;
}

void Player::integrateKeys(double dt)
{
	// steering
	
	if (keys[kSteerWheelLeft])
	{
		car.mSteerWheelAngle += kSteerSpeed * dt;
		if (car.mSteerWheelAngle > kMaxSteerAngle)
			car.mSteerWheelAngle = kMaxSteerAngle;
	}
	
	if (keys[kSteerWheelRight])
	{
		car.mSteerWheelAngle -= kSteerSpeed * dt;
		if (car.mSteerWheelAngle < -kMaxSteerAngle)
			car.mSteerWheelAngle = -kMaxSteerAngle;
	}
	
	if (!(keys[kSteerWheelLeft] || keys[kSteerWheelRight]))
	{
		reduce_by_at_most(car.mSteerWheelAngle, kSteerReturnSpeed*dt);
	}
	
	// throttle

	if (keys[kThrottleMore])
	{
		car.mThrottle += kThrottleSpeed * dt;
		if (car.mThrottle > 1.0)
			car.mThrottle = 1.0;
	}
	
	if (keys[kThrottleLess])
	{
		car.mThrottle -= kThrottleSpeed * dt;
		if (car.mThrottle < -1.0)
			car.mThrottle = -1.0;
	}
	
	if (!(keys[kThrottleMore] || keys[kThrottleLess]))
	{
		reduce_by_at_most(car.mThrottle, kThrottleReturnSpeed*dt);
	}
	
	// brake

	if (keys[kBrakeMore])
	{
		car.mBrake += kBrakeSpeed * dt;
		if (car.mBrake > 1.0)
			car.mBrake = 1.0;
	}
	
	if (keys[kBrakeLess])
	{
		car.mBrake -= kBrakeSpeed * dt;
		if (car.mBrake < 0.0)
			car.mBrake = 0.0;
	}
	
	if (!(keys[kBrakeMore] || keys[kBrakeLess]))
	{
		reduce_by_at_most(car.mBrake, kBrakeReturnSpeed*dt);
	}
}

void Player::reset(const Tuple& lineA, const Tuple& lineB, const Player *winner)
{
	const Tuple vec1 = car.getLocation_wc()*kScale - lineA;
	const Tuple vec2 = lineB - lineA;
    
    prevLap = curLap = vec2.Angle(vec1);
    
    lapFlag = 0;
	laps = 0;
    revolutions = 0;
    
	bestLapTime = INFINITY;
    courseTime = 0;
	
	if (this != winner)
	{
		computePathLocation();
		pathIndex += (winner->pathIndex / pathSize) * pathSize;
		while (winner->pathIndex - pathIndex > pathSize/2)
			pathIndex += pathSize;
		while (pathIndex - winner->pathIndex > pathSize/2)
			pathIndex -= pathSize;
	}
}
