/*
	Copyright (C) 2005 by Ruben Zilibowitz <rzilibowitz@users.sourceforge.net>
	Part of the Toy Cars Project http://toycars.sourceforge.net

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License.
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY.

	See the COPYING file for more details.
*/

/*
 *  ScSprite.cpp
 *  Scoobie
 *
 *  Created by Ruben Zilibowitz on 26/11/04.
 *
 */

#include "ScSprite.h"

//
// constructor - creates an empty sprite of given size
//
ScSprite::ScSprite(const ScRect& inBounds)
	: ScPanel(inBounds), isMaster(false), numFrames(0), curFrame(0), frames(NULL)
{
}

//
// constructor - creates an empty sprite of given size
//
ScSprite::ScSprite(ScFloat width, ScFloat height)
	: ScPanel(width, height), isMaster(false), numFrames(0), curFrame(0), frames(NULL)
{
}

//
// constructor - creates and loads a master sprite
//
ScSprite::ScSprite(int framewidth, int frameheight, short border, const ScPixelData& pixelData)
	: ScPanel(framewidth - 2*border, frameheight - 2*border)
{
	// load sprite frames
	load(framewidth, frameheight, border, pixelData);
	
	// init curFrame
	curFrame = 0;
	
	// this is a master sprite
	isMaster = true;
}

//
// copy constructor - will use the frame data of another sprite.
// Should use when two sprites share same frame set.
//
ScSprite::ScSprite(const ScSprite& copySprite)
	:	ScPanel(copySprite.bounds),
		numFrames(copySprite.numFrames),
		frames(copySprite.frames)
{
	// init the current frame
	curFrame = 0;
	
	// this is NOT a master sprite
	isMaster = false;
}

//
// destructor - free memory and texture if we are a master sprite
//
ScSprite::~ScSprite()
{
	if (isMaster)
	{
		// free frame images
		for (int i = 0; i < numFrames; i++)
			delete frames[i];
		delete [] frames;
	}
}

//
// doDraw - draws current frame at location given by bounds
//
void ScSprite::doDraw()
{
	glColor4ub(255, 255, 255, 255);
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);
	frames[curFrame]->draw(bounds.left, bounds.bottom);
}

//
// load - Loads sprite into textures. Pixel data in pixels is assumed
// to be at least framewidth by frameheight in size. Frames are loaded from the
// pixel data by starting at 0,0 and moving to the right, then moving upwards
// (ie rows first). Pixels are assumed to be in "Scoobie" coordinates where
// 0,0 is the bottom left of the image. Therefore an image file should usually
// be flipped vertically either upon loading it, or in the original file itself.
//
void ScSprite::load(int framewidth, int frameheight, short border, const ScPixelData& pixelData)
{
	int xframes = pixelData.width / framewidth;
	int yframes = pixelData.height / frameheight;
	int i, j, k;
	long offset;
	ScPixelData framePixelData;
	
	// setup frame pixel data struct
	framePixelData = pixelData;
	framePixelData.width = framewidth;
	framePixelData.height = frameheight;
	
	// calculate numFrames
	numFrames = xframes*yframes;
	
	// allocate array
	frames = new ScImage * [numFrames];
	
	// init frame counter
	k = 0;
	
	// step through frames left to right then bottom to top
	// load each one into texture memory
	for (j = 0; j < yframes; j++)
	{
		for (i = 0; i < xframes; i++)
		{
			offset = (j * frameheight * pixelData.pitch) + (i * framewidth * pixelData.bytes_pp);
			framePixelData.pixels = pixelData.pixels + offset;
			
			frames[k] = new ScImage(border, framePixelData);
			k++;
		}
	}
}
