/*
	Copyright (C) 2005 by Ruben Zilibowitz <rzilibowitz@users.sourceforge.net>
	Part of the Toy Cars Project http://toycars.sourceforge.net

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License.
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY.

	See the COPYING file for more details.
*/

/*
 *  Created by Ruben Zilibowitz on 31/12/04.
 */

#include "ToyCar.h"
#include "ScException.h"
#include "GameConstants.h"

#include "SDL.h"
#include "SDL_image.h"
#include <cstdlib>

// Sprite Constants

const char kCarSpriteFile[] = "images/buggy.png";
const int kCarFrameWidth = 75;
const int kCarFrameHeight = 75;
const ScRect kCarBounds = {0, 0, kCarFrameWidth, kCarFrameHeight};

// car geometry - remember polygons are counter-clockwise!

const Tuple kChassis[] = {  Tuple(3-38,38-48)/kScale, Tuple(8-38,38-54)/kScale, Tuple(67-38,38-54)/kScale,
                            Tuple(71-38,38-48)/kScale, Tuple(71-38,38-28)/kScale, Tuple(67-38,38-22)/kScale,
                            Tuple(8-38,38-22)/kScale, Tuple(3-38,38-28)/kScale };

const Tuple kRearWheels[] = {   Tuple(10-38,38-21)/kScale, Tuple(10-38,38-55)/kScale, Tuple(13-38,38-57)/kScale,
                                Tuple(25-38,38-57)/kScale, Tuple(28-38,38-55)/kScale, Tuple(28-38,38-21)/kScale,
                                Tuple(25-38,38-19)/kScale, Tuple(13-38,38-19)/kScale };

const Convex kCarGeom[] = { Convex(kChassis, 8), Convex(kRearWheels, 8) };

//
// Car properties
//

const double kWidth = 32 / kScale;              // m
const double kFrontAxel = 22 / kScale;          // m
const double kRearAxel = 19 / kScale;           // m
const double kFrontWheelInset = 6 / kScale;     // m
const double kFrontWheelRadius = 7 / kScale;	// m
const double kRearWheelRadius = 8 / kScale;	// m

const double kCarMass = 580;			// kg
const double kCarInertia = 300;         // kg m^2
const double kFrontWheelInertia = 1.3;	// kg m^2
const double kRearWheelInertia = 1.3;	// kg m^2

const double kCarDrag = 4;				// units?
const double kRollingResistance = 0.06;	// Nm

#pragma mark class ToyCar

///--------------------------------------------------------------------------------------
//  ToyCar::ToyCar
///--------------------------------------------------------------------------------------

ToyCar::ToyCar( const char * spritefilename,
                short inFramewidth,
                short inFrameheight,
                
                const char *wheelspritefilename,
                short wheelframewidth,
                short wheelframeheight,
                char inWheelLayer,
                
                double inMass,
                double inInertia,

                Tuple inStartLocation,
                double inStartOrientation,

                double inFrontAxel,
                double inRearAxel,
                double inAxelWidth,
                double inPinInset,

                double inFrontTyreRadius,
                double inRearTyreRadius,
                double inFrontWheelInertia,
                double inRearWheelInertia,

                PacejkaCurve& inFrontWheelSlipRatioCurve,
                PacejkaCurve& inFrontWheelSlipAngleCurve,
                PacejkaCurve& inRearWheelSlipRatioCurve,
                PacejkaCurve& inRearWheelSlipAngleCurve,

                const Geom& inGeom,

                double inDrag,
                double inRollingResistance)
    : Vehicle(
        inMass,           // mass
        inInertia,        // inertia
		
        inStartLocation,    // start location given in metres
        inStartOrientation, // start orientation given in radians
		
        inFrontAxel,         // distance to front axel
        inRearAxel,          // distance to rear axel
        inAxelWidth,        // distance between wheels (front and back)
        inPinInset,         // distance between hinge joint and contact patch
		
        inFrontTyreRadius,      // front wheel radius
        inRearTyreRadius,       // rear wheel radius
        inFrontWheelInertia,    // front wheel inertia
        inRearWheelInertia,     // rear wheel inertia
		
		*(frontSlipRatioCurve = &inFrontWheelSlipRatioCurve),     // front wheel slip ratio curve
		*(frontSlipAngleCurve = &inFrontWheelSlipAngleCurve),     // front wheel slip angle curve
		*(rearSlipRatioCurve = &inRearWheelSlipRatioCurve),      // rear wheel slip ratio curve
		*(rearSlipAngleCurve = &inRearWheelSlipAngleCurve),      // rear wheel slip angle curve
		
        inGeom,  // geometry
		
        inDrag,               // drag
        inRollingResistance),    // rolling resistance
        
		ScSprite(inFramewidth, inFrameheight),	// constructor for ScSprite base class
		
        wheelLayer(inWheelLayer)

{
	SDL_Surface* imgSurface;
	ScPixelData pixelData;
	
	// this is a master sprite
	isMaster = true;
    
	// create image surface
	imgSurface = IMG_Load(spritefilename);
	if(!imgSurface)
	{
		ScThrowErr("Could not load image file");
	}
	
	// setup the pixel data struct
	pixelData.pitch = imgSurface->pitch;
	pixelData.bytes_pp = imgSurface->format->BytesPerPixel;
	pixelData.internalformat = GL_RGBA;
	pixelData.width = imgSurface->w;
	pixelData.height = imgSurface->h;
	pixelData.format = GL_RGBA;
	pixelData.type = GL_UNSIGNED_BYTE;
	pixelData.pixels = (const char*)(imgSurface->pixels);
		
	// load the sprite frames
	load(inFramewidth, inFrameheight, 0, pixelData);
	
	// free the image surface
	SDL_FreeSurface(imgSurface);
	
	// create the wheel sprites
	leftWheelSprite = new WheelSprite(this, wheelspritefilename, wheelframewidth, wheelframeheight, &mWheel_FL, 0.0);
	rightWheelSprite = new WheelSprite(*leftWheelSprite, this, &mWheel_FR, kPi);
}

///--------------------------------------------------------------------------------------
//  ToyCar::ToyCar
///--------------------------------------------------------------------------------------

ToyCar::ToyCar(Tuple inStartLocation, double inStartOrientation)
    : Vehicle(
        kCarMass,           // mass
        kCarInertia,        // inertia
		
        inStartLocation,    // start location given in metres
        inStartOrientation, // start orientation given in radians
		
        kFrontAxel,         // distance to front axel
        kRearAxel,          // distance to rear axel
        kWidth,             // distance between wheels (front and back)
        kFrontWheelInset,   // distance between hinge joint and contact patch
		
        kFrontWheelRadius,  // front wheel radius
        kRearWheelRadius,   // rear wheel radius
        kFrontWheelInertia, // front wheel inertia
        kRearWheelInertia,  // rear wheel inertia
		
		*(frontSlipRatioCurve = new PacejkaCurve(3.0, 1.4, 7.0, -0.6)),		// front wheel slip ratio curve
		*(frontSlipAngleCurve = new PacejkaCurve(3.0, 1.35, 13.5, -0.4)),	// front wheel slip angle curve
		*(rearSlipRatioCurve = new PacejkaCurve(3.0, 1.4, 7.0, -0.6)),		// rear wheel slip ratio curve
		*(rearSlipAngleCurve = new PacejkaCurve(3.0, 1.35, 13.5, -0.4)),	// rear wheel slip angle curve
		
        Geom(kCarGeom, 2),  // geometry
		
        kCarDrag,           // drag
        kRollingResistance),// rolling resistance
        
		ScSprite(kCarBounds),	// constructor for ScSprite base class
		
        wheelLayer(0)
{
	SDL_Surface* imgSurface;
	ScPixelData pixelData;
	
	// this is a master sprite
	isMaster = true;

	// create image surface
	imgSurface = IMG_Load(kCarSpriteFile);
	if(!imgSurface)
	{
		ScThrowErr("Could not load image file");
	}
	
	// setup the pixel data struct
	pixelData.pitch = imgSurface->pitch;
	pixelData.bytes_pp = imgSurface->format->BytesPerPixel;
	pixelData.internalformat = GL_RGBA;
	pixelData.width = imgSurface->w;
	pixelData.height = imgSurface->h;
	pixelData.format = GL_RGBA;
	pixelData.type = GL_UNSIGNED_BYTE;
	pixelData.pixels = (const char*)(imgSurface->pixels);
		
	// load the sprite frames
	load(kCarFrameWidth, kCarFrameHeight, 0, pixelData);
	
	// free the image surface
	SDL_FreeSurface(imgSurface);
	
	// create the wheel sprites
	leftWheelSprite = new WheelSprite(this, &mWheel_FL, 0.0);
	rightWheelSprite = new WheelSprite(*leftWheelSprite, this, &mWheel_FR, kPi);
}

///--------------------------------------------------------------------------------------
//  ToyCar::ToyCar
//  copy constructor
///--------------------------------------------------------------------------------------

ToyCar::ToyCar(const ToyCar& copyCar, Tuple inStartLocation, double inStartOrientation)
	:   Vehicle(copyCar, inStartLocation, inStartOrientation),
		ScSprite(copyCar),	// clone sprite
        wheelLayer(copyCar.wheelLayer)
{
	leftWheelSprite = new WheelSprite(*copyCar.leftWheelSprite, this, &mWheel_FL, 0.0);
	rightWheelSprite = new WheelSprite(*copyCar.leftWheelSprite, this, &mWheel_FR, kPi);
}

///--------------------------------------------------------------------------------------
//  ToyCar::~ToyCar
///--------------------------------------------------------------------------------------

ToyCar::~ToyCar()
{
	if (isMaster)
	{
		delete frontSlipRatioCurve;
		delete frontSlipAngleCurve;
		delete rearSlipRatioCurve;
		delete rearSlipAngleCurve;
	}
	
	delete leftWheelSprite;
	delete rightWheelSprite;
}

///--------------------------------------------------------------------------------------
//  ToyCar::process
///--------------------------------------------------------------------------------------

void ToyCar::process(ScFloat dt)
{
	centreAt(mLocation_wc.x * kScale, mLocation_wc.y * kScale);
    curFrame = iround(mOrientation_wc * kInvTwoPi * numFrames) % numFrames;
	while (curFrame < 0)	curFrame += numFrames;
    
    leftWheelSprite->process(dt);
    rightWheelSprite->process(dt);
}

///--------------------------------------------------------------------------------------
//  ToyCar::doDraw
///--------------------------------------------------------------------------------------

void ToyCar::doDraw()
{
    if (wheelLayer == 0) {
        leftWheelSprite->doDraw();
        rightWheelSprite->doDraw();
        ScSprite::doDraw();
    }
    else { // wheelLayer == 1
        ScSprite::doDraw();
        leftWheelSprite->doDraw();
        rightWheelSprite->doDraw();
    }
}
