/*
	Copyright (C) 2005 by Ruben Zilibowitz <rzilibowitz@users.sourceforge.net>
	Part of the Toy Cars Project http://toycars.sourceforge.net

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License.
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY.

	See the COPYING file for more details.
*/

/*
 *  Created by Ruben Zilibowitz on 31/12/04.
 */

#include "Vehicle.h"
#include "GameConstants.h"

///--------------------------------------------------------------------------------------
//  Vehicle::Vehicle
///--------------------------------------------------------------------------------------

Vehicle::Vehicle(   double inMass,
                    double inInertia,
					
                    Tuple inStartLocation,
                    double inStartOrientation,
					
                    double inFrontAxel,
                    double inRearAxel,
                    double inAxelWidth,
                    double inPinInset,
					
                    double inFrontTyreRadius,
                    double inRearTyreRadius,
                    double inFrontWheelInertia,
                    double inRearWheelInertia,
					
					PacejkaCurve& inFrontWheelSlipRatioCurve,
					PacejkaCurve& inFrontWheelSlipAngleCurve,
					PacejkaCurve& inRearWheelSlipRatioCurve,
					PacejkaCurve& inRearWheelSlipAngleCurve,
					
                    const Geom& inGeom,
					
                    double inDrag,
                    double inRollingResistance)
    : RigidBody(
        inMass,             // mass
        inInertia,          // inertia
        inStartLocation,    // location
        Tuple(0,0),         // velocity
        inStartOrientation, // orientation
        0,                  // angular velocity
        inGeom              // geometry
        ),
        
        mSteerWheelAngle(0.0),
        mThrottle(0.0),
        mBrake(0.0),
        
        mFrontAxel(inFrontAxel),
        mRearAxel(inRearAxel),
        mWheelbase(mFrontAxel + mRearAxel),
        mAxelWidth(inAxelWidth),
        
        mDragConstant(inDrag),
        
        mWheel_FL(*this, inFrontWheelSlipRatioCurve, inFrontWheelSlipAngleCurve, Tuple(mFrontAxel, mAxelWidth/2 - inPinInset), inFrontTyreRadius, inPinInset, 0, inFrontWheelInertia, 4000, inRollingResistance),
        mWheel_FR(*this, inFrontWheelSlipRatioCurve, inFrontWheelSlipAngleCurve, Tuple(mFrontAxel, -mAxelWidth/2 + inPinInset), inFrontTyreRadius, -inPinInset, 0, inFrontWheelInertia, 4000, inRollingResistance),
        mWheel_BL(*this, inRearWheelSlipRatioCurve, inRearWheelSlipAngleCurve, Tuple(-mRearAxel, mAxelWidth/2), inRearTyreRadius, 0, 0, inRearWheelInertia, 4000, inRollingResistance),
        mWheel_BR(*this, inRearWheelSlipRatioCurve, inRearWheelSlipAngleCurve, Tuple(-mRearAxel, -mAxelWidth/2), inRearTyreRadius, 0, 0, inRearWheelInertia, 4000, inRollingResistance)
{
    // calculate static weight distribution
    
    mWheel_FL.mLoad = 0.5 * mMass * 9.8 * mRearAxel / mWheelbase;
    mWheel_FR.mLoad = 0.5 * mMass * 9.8 * mRearAxel / mWheelbase;
    mWheel_BL.mLoad = 0.5 * mMass * 9.8 * mFrontAxel / mWheelbase;
    mWheel_BR.mLoad = 0.5 * mMass * 9.8 * mFrontAxel / mWheelbase;
	
	mEngineBrake = 1500.0;
}

///--------------------------------------------------------------------------------------
//  Vehicle::Vehicle
///--------------------------------------------------------------------------------------

Vehicle::Vehicle(const Vehicle& copyVehicle, Tuple inStartLocation, double inStartOrientation)
    : RigidBody(
        copyVehicle.mMass,             // mass
        copyVehicle.mInertia,          // inertia
        inStartLocation,    // location
        Tuple(0,0),         // velocity
        inStartOrientation, // orientation
        0,                  // angular velocity
        Geom(copyVehicle.mGeom) // geometry
        ),
        
        mSteerWheelAngle(0.0),
        mThrottle(0.0),
        mBrake(0.0),
        
        mFrontAxel(copyVehicle.mFrontAxel),
        mRearAxel(copyVehicle.mRearAxel),
        mWheelbase(copyVehicle.mWheelbase),
        mAxelWidth(copyVehicle.mAxelWidth),
        
        mDragConstant(copyVehicle.mDragConstant),
        
        mWheel_FL(*this, copyVehicle.mWheel_FL),
        mWheel_FR(*this, copyVehicle.mWheel_FR),
        mWheel_BL(*this, copyVehicle.mWheel_BL),
        mWheel_BR(*this, copyVehicle.mWheel_BR)
{
    // calculate static weight distribution
    
    mWheel_FL.mLoad = 0.5 * mMass * 9.8 * mRearAxel / mWheelbase;
    mWheel_FR.mLoad = 0.5 * mMass * 9.8 * mRearAxel / mWheelbase;
    mWheel_BL.mLoad = 0.5 * mMass * 9.8 * mFrontAxel / mWheelbase;
    mWheel_BR.mLoad = 0.5 * mMass * 9.8 * mFrontAxel / mWheelbase;
	
	mEngineBrake = 1500.0;
}

///--------------------------------------------------------------------------------------
//  Vehicle::compute_force_and_torque
///--------------------------------------------------------------------------------------

void Vehicle::compute_force_and_torque()
{	
	// transform velocity into local coordinates
	
    mVelocity = mNorm.toLocal(mVelocity_wc);
    
	// calculate wheel angles and positions

    mWheel_FL.setOrientation(mSteerWheelAngle);
    mWheel_FR.setOrientation(mSteerWheelAngle);
    	
	// drag and rolling resistance
	
    mDrag = -(mDragConstant * mVelocity.Length()) * mVelocity; // nb: a sqrt is being used here which could be avoided
	
	// calculate vertical load on tyres
	// NYI
	
	// apply wheel brakes
	
    mWheel_FL.mBrake = mBrake;
    mWheel_FR.mBrake = mBrake;
    mWheel_BL.mBrake = mBrake;
    mWheel_BR.mBrake = mBrake;
	
	// calculate tyre forces and torques
	
    mWheel_FL.compute_force_and_torque();
    mWheel_FR.compute_force_and_torque();
    mWheel_BL.compute_force_and_torque();
    mWheel_BR.compute_force_and_torque();
    
	// calculate rpm of engine from wheel angular velocities
	
	mRPM = 0.5 * (mWheel_BL.mAngularVelocity + mWheel_BR.mAngularVelocity) * (60.0 * kInvTwoPi);
	
	// lookup engine torque from rpm-torque curve
	
	mDriveTorque = mThrottle * engine_torque_curve(sign(mThrottle)*mRPM);
	mDriveTorque -= sign_smooth(mRPM, 25.0) * (1-fabs(mThrottle)) * mEngineBrake;
	
	// apply engine torque to wheels via differential gear train
	// linear equation depends on the moments of the components in the gear train
	
	mWheel_BL.mTotalTorque = 0.9*mWheel_BL.mTractionTorque - 0.05*mWheel_BR.mTractionTorque + 0.085*mDriveTorque;
	mWheel_BR.mTotalTorque = -0.05*mWheel_BL.mTractionTorque + 0.9*mWheel_BR.mTractionTorque + 0.085*mDriveTorque;
	
	// free rolling wheels
	
	mWheel_FL.mTotalTorque = mWheel_FL.mTractionTorque;
	mWheel_FR.mTotalTorque = mWheel_FR.mTractionTorque;
	
    // calculate angular accel.
    
    mWheel_FL.mAngularAcceleration = mWheel_FL.mTotalTorque * mWheel_FL.mInvInertia;
    mWheel_FR.mAngularAcceleration = mWheel_FR.mTotalTorque * mWheel_FR.mInvInertia;
    mWheel_BL.mAngularAcceleration = mWheel_BL.mTotalTorque * mWheel_BL.mInvInertia;
    mWheel_BR.mAngularAcceleration = mWheel_BR.mTotalTorque * mWheel_BR.mInvInertia;
    
    // sum forces
    
    mTotalForce = mWheel_FL.mForce_vc + mWheel_FR.mForce_vc +
                  mWheel_BL.mForce_vc + mWheel_BR.mForce_vc + mDrag;
	
    // transform into world coords
    
    mTotalForce_wc = mNorm.fromLocal(mTotalForce);
	
	// sum torques
	
    mTotalTorque = mWheel_FL.mVehicleTorque + mWheel_FR.mVehicleTorque +
                   mWheel_BL.mVehicleTorque + mWheel_BR.mVehicleTorque;
}

///--------------------------------------------------------------------------------------
//  Vehicle::engine_torque_curve
///--------------------------------------------------------------------------------------

double Vehicle::engine_torque_curve(double rpm)
{
/*
    // Mabuchi RS-540SH/SF-5045 12.0V torque curve
    
	// max torque output of engine 230 Nm
	// max rpm about 17692
	
	if (rpm < 0.0)
		return 230;
	else
		return 230 - rpm*(0.0130);
*/
	
	if (rpm < 0.0)
		return 4000;
	else
		return 4000 - rpm*(0.5);
}

///--------------------------------------------------------------------------------------
//  Vehicle::step - method used is 2nd order Runge-Kutta
///--------------------------------------------------------------------------------------

void Vehicle::rk2_step(double dt)
{
		// declare these variables as static since
		// this function is called frequently
	static Tuple saveLoc;
	static double saveOrient;
	static Tuple saveVel;
	static double saveAngVel;
	static double saveAngVel_FL;
	static double saveAngVel_FR;
	static double saveAngVel_BL;
	static double saveAngVel_BR;
	
	saveLoc = mLocation_wc;
	saveOrient = mOrientation_wc;
	saveVel = mVelocity_wc;
	saveAngVel = mAngularVelocity;
	saveAngVel_FL = mWheel_FL.mAngularVelocity;
	saveAngVel_FR = mWheel_FR.mAngularVelocity;
	saveAngVel_BL = mWheel_BL.mAngularVelocity;
	saveAngVel_BR = mWheel_BR.mAngularVelocity;
	
	dt *= 0.5;
	
    mLocation_wc += dt * mVelocity_wc;
    mOrientation_wc += dt * mAngularVelocity;
	
    mVelocity_wc += dt * mAcceleration_wc;
    mAngularVelocity += dt * mAngularAcceleration;
    
    mWheel_FL.mAngularVelocity += dt * mWheel_FL.mAngularAcceleration;
    mWheel_FR.mAngularVelocity += dt * mWheel_FR.mAngularAcceleration;
    mWheel_BL.mAngularVelocity += dt * mWheel_BL.mAngularAcceleration;
    mWheel_BR.mAngularVelocity += dt * mWheel_BR.mAngularAcceleration;
    
    if (mOrientation_wc < 0)
        mOrientation_wc += kTwoPi;
    else if (mOrientation_wc >= kTwoPi)
        mOrientation_wc -= kTwoPi;
    
	mNorm.x = -sin(mOrientation_wc);
    mNorm.y = cos(mOrientation_wc);
    
    compute_force_and_torque();
    mAcceleration_wc = mInvMass * mTotalForce_wc;
	mAngularAcceleration = mInvInertia * mTotalTorque;
    
	dt *= 2.0;
	
    mLocation_wc = saveLoc + (dt * mVelocity_wc);
    mOrientation_wc = saveOrient + (dt * mAngularVelocity);
	
    mVelocity_wc = saveVel + (dt * mAcceleration_wc);
    mAngularVelocity = saveAngVel + (dt * mAngularAcceleration);
    
    mWheel_FL.mAngularVelocity = saveAngVel_FL + (dt * mWheel_FL.mAngularAcceleration);
    mWheel_FR.mAngularVelocity = saveAngVel_FR + (dt * mWheel_FR.mAngularAcceleration);
    mWheel_BL.mAngularVelocity = saveAngVel_BL + (dt * mWheel_BL.mAngularAcceleration);
    mWheel_BR.mAngularVelocity = saveAngVel_BR + (dt * mWheel_BR.mAngularAcceleration);
    
    if (mOrientation_wc < 0)
        mOrientation_wc += kTwoPi;
    else if (mOrientation_wc >= kTwoPi)
        mOrientation_wc -= kTwoPi;
    
	mNorm.x = -sin(mOrientation_wc);
    mNorm.y = cos(mOrientation_wc);
    
    compute_force_and_torque();
    mAcceleration_wc = mInvMass * mTotalForce_wc;
	mAngularAcceleration = mInvInertia * mTotalTorque;
}

void Vehicle::kill()
{
	mVelocity_wc = Tuple(0,0);
	mAngularVelocity = 0;
	mSteerWheelAngle = 0;
	mThrottle = 0;
	mBrake = 0;
}
