/*
	Copyright (C) 2005 by Ruben Zilibowitz <rzilibowitz@users.sourceforge.net>
	Part of the Toy Cars Project http://toycars.sourceforge.net

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License.
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY.

	See the COPYING file for more details.
*/

/*
 *  Created by Ruben Zilibowitz on 31/12/04.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#endif

#if defined(__APPLE__) && defined(__MACH__)
#include <OpenGL/gl.h>	// Header File For The OpenGL32 Library
#else
#include <GL/gl.h>	// Header File For The OpenGL32 Library
#endif

#include "SDL.h"
#include "SDL_image.h"

#include <cstdio>
#include <cstring>
#include <cstdlib>
#include <ctime>
#include <fcntl.h>
#include <errno.h>

#include "ScException.h"
#include "CarGame.h"
#include "HotPotatoGame.h"
#include "KnockOutGame.h"
#include "GetSeconds.h"
#include "GameConstants.h"
#include "tinyxml.h"
#include "glf.h"
#include "Menu.h"
#include "TcPreferences.h"
#include "RobotPlayer.h"
/*
#include <pthread.h>
#include "alut.h"
#include "ogg_stream.h"
*/

#include "SDL_mixer.h"

//
// Constants

const char kLoadingImageFileName[] = "images/title.png";	// path to loading image from within Data dir
const char kTitleSongFileName[] = "sounds/Motor_Car_Song.ogg";
//const bool kUseVBLSynch = false;	// synchronise with the screen retraces?

//
// Globals

char gDataDir[256] = DATA_DIR;		// Path to the Data dir relative to the executable
CarGame* gGameP = NULL;
unsigned int gNumPlayers;	// The number of players
unsigned int gNumHumanPlayers;
TcPreferences *gPreferences;
Mix_Music *gMusic = NULL;

//
// Game Mode enum

enum GameMode
{
	GameMode_TimeTrial,
	GameMode_HotPotato,
	GameMode_KnockEmOut
} gGameMode;

//
// Functions

void *streamMusic(void *arg);

void PixelDataForSurface(const SDL_Surface* src, ScPixelData& dst)
{
	dst.pitch = src->pitch;
	dst.bytes_pp = src->format->BytesPerPixel;
	dst.width = src->w;
	dst.height = src->h;
	
	switch (dst.bytes_pp)
	{
		case 4:
			dst.format = GL_RGBA;
			dst.type = GL_UNSIGNED_BYTE;
			dst.internalformat = GL_RGBA;
			break;
		
		case 3:
			dst.format = GL_RGB;
			dst.type = GL_UNSIGNED_BYTE;
			dst.internalformat = GL_RGB;
			break;
		
		default:
			ScThrowErr("Only 32 bit or 24 bit is supported at present");
			break;
	}
	
	dst.pixels = (const char*)src->pixels;
}

void DrawSplash()
{
	SDL_Surface* img;
	char string64[64];
	//float minx, miny, maxx, maxy;
    
	// Load the image
	img = IMG_Load(kLoadingImageFileName);
	if (img == NULL)
		ScThrowErr("Could not load splash screen image file");
	
	// Draw the image
	glPixelStorei(GL_UNPACK_ALIGNMENT, 4);
	glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
    glLoadIdentity();
	glOrtho(0, gPreferences->getScreenWidth(), 0, gPreferences->getScreenHeight(), -1, 1);
	glPixelZoom(1, -1);
	glRasterPos2f(gPreferences->getScreenWidth()*0.5 - img->w*0.5, gPreferences->getScreenHeight()*0.5 + img->h*0.5);
	glClear(GL_COLOR_BUFFER_BIT);
	glDrawPixels(img->w, img->h, GL_RGB, GL_UNSIGNED_BYTE, img->pixels);
	
	// Draw version number
	sprintf(string64, "v. %s", VERSION);
	glColor4f(1.0f, 0.0f, 0.0f, 1.0f);
	//glfGetStringBounds(string64, &minx, &miny, &maxx, &maxy);
	glTranslatef(gPreferences->getScreenWidth() - 96, 20, 0);
	glScalef(12, 12, 1);
	glfDrawSolidString(string64);
    
	// Free sdl image and swap gl buffers
	SDL_FreeSurface(img);
	//SDL_GL_SwapBuffers();
}

void DrawLoading()
{
    // Draw loading text
	glColor4f(1.0f, 0.0f, 0.0f, 1.0f);
    glLoadIdentity();
	glOrtho(0, gPreferences->getScreenWidth(), 0, gPreferences->getScreenHeight(), -1, 1);
	glScalef(24, 24, 1);
	glTranslatef(1, 1, 0);
	glfDrawSolidString("Loading...");
    SDL_GL_SwapBuffers();
}

void DrawPaused()
{
    // Draw loading text
	glColor4f(1.0f, 0.0f, 0.0f, 1.0f);
    glLoadIdentity();
	glOrtho(0, gPreferences->getScreenWidth(), 0, gPreferences->getScreenHeight(), -1, 1);
	glScalef(12, 12, 1);
	glTranslatef(1, 1, 0);
	glfDrawSolidString("Paused (press p to unpause)...");
    SDL_GL_SwapBuffers();
}

void ShowCredits()
{
    // Draw loading text
	glColor4f(1.0f, 0.0f, 0.0f, 1.0f);
    glLoadIdentity();
	glOrtho(0, gPreferences->getScreenWidth(), 0, gPreferences->getScreenHeight(), -1, 1);
	glTranslatef(10, gPreferences->getScreenHeight()*0.5, 0);
	glScalef(8, 8, 1);
	glfDrawSolidString("Ruben Zilibowitz <rzilibowitz@users.sourceforge.net> - Main Developer");
	glTranslatef(0, -2.5, 0);
	glfDrawSolidString("James Englund with Crash Symphony (http://www.crashsymphony.com/) - The Motor Car Song");
    glLoadIdentity();
	glOrtho(0, gPreferences->getScreenWidth(), 0, gPreferences->getScreenHeight(), -1, 1);
	glScalef(10, 10, 1);
	glTranslatef(1, 1, 0);
	glfDrawSolidString("Press any key to continue...");
    SDL_GL_SwapBuffers();
	
	SDL_Event event;
	while (1)
	{
		if (SDL_WaitEvent(&event) == 0) {
			printf("error waiting for events\n");
			exit(EXIT_FAILURE);
		}
		if (event.type == SDL_KEYUP)
			break;
	}
}

/*
* This little groovy formula I cooked up in Mathematica. It returns the shortest line
* distance from the point (x,y) to the ellipse defined by (x/a)^2 + (y/b)^2 == 1.
*/
/*
double ellipsePerpDist(double x, double y, double a, double b)
{
	double x2 = x*x;
	double y2 = y*y;
	double a4 = a*a*a*a;	
	double a5 = a*a*a*a*a;	
	double a6 = a*a*a*a*a*a;	
	double b4 = b*b*b*b;	
	double b5 = b*b*b*b*b;	
	double b6 = b*b*b*b*b*b;
	
	double AA = sqr(b4*x2 + a4*y2);
	double BB = a*b5*x2 + a5*b*y2;
	double CC = sqrt(b6*x2 + (a6 - sqr(a*a - b*b)*x2)*y2);
	double DD = sqrt(b4*x2 + a4*y2);
	double EE = b6*x2 + a6*y2;
	
	return (AA - BB*CC)/(DD*EE);
}
*/
ScTilemap* loadTileset(const char* tilesetFilename)
{
	char path[256] = "tilesets/";
	char filename[256];
	
	strcat(path, tilesetFilename);
	strcat(path, "/");
	strcpy(filename, path);
	strcat(filename, tilesetFilename);
	strcat(filename, ".xml");
	
	// Load the xml file
	TiXmlDocument *xmlDoc = new TiXmlDocument(filename);
	if (!xmlDoc->LoadFile())
	{
		printf("could not find file %s\n", filename);
		ScThrowErr("xml file not found");
	}
	
	// Do the document parsing here
	TiXmlElement *xTileset = xmlDoc->FirstChildElement("tileset");
	if (!xTileset)
	{
		xmlDoc->Clear();
		delete xmlDoc;
		ScThrowErr("xml tileset element not found");
	}
	
	TiXmlHandle tilesetHandle(xTileset);
	
	TiXmlText *xImage = tilesetHandle.FirstChild("image").FirstChild().Text();
	if (!xImage)
	{
		xmlDoc->Clear();
		delete xmlDoc;
		ScThrowErr("xml image element not found");
	}
	
	TiXmlText *xTilewidth = tilesetHandle.FirstChild("tilewidth").FirstChild().Text();
	if (!xTilewidth)
	{
		xmlDoc->Clear();
		delete xmlDoc;
		ScThrowErr("xml tilewidth element not found");
	}
	
	TiXmlText *xTileheight = tilesetHandle.FirstChild("tileheight").FirstChild().Text();
	if (!xTileheight)
	{
		xmlDoc->Clear();
		delete xmlDoc;
		ScThrowErr("xml tileheight element not found");
	}
	
	ScPixelData pixelData;
	SDL_Surface* surface;
	ScTilemap* tilemap;
	int tileWidth = strtol(xTilewidth->Value(), (char **)NULL, 10);
	int tileHeight = strtol(xTileheight->Value(), (char **)NULL, 10);
	
	if (tileWidth <= 0)
		ScThrowErr("tilewidth is invalid string");
	if (tileHeight <= 0)
		ScThrowErr("tileheight is invalid string");
	
	// create the tile map
	strcpy(filename, path);
	strcat(filename, xImage->Value());
	surface = IMG_Load(filename);
	if (surface == NULL)
		ScThrowErr("Could not load tiles image file");
	PixelDataForSurface(surface, pixelData);
	tilemap = new ScTilemap(pixelData, tileWidth, tileWidth, 0);
	SDL_FreeSurface(surface);
	
	printf("successfully created tileset from %s\n", path);
	
	return tilemap;
}

void readPointsFromString(const char* str, Tuple* points, short& count)
{
    char *word, *last;
    char *temp = strdup(str);
    
    for (   word = strtok_r(temp, ";", &last), count = 0;
            word;
            word = strtok_r(NULL, ";", &last), count++)
    {
        points[count].x = atof(word);
        if ((word = strchr(word, ',')) == NULL)
            ScThrowErr("Badly formatted points list");
        points[count].y = atof(word + 1);
    }
    
    free(temp);
}

char* chooseToyCar(int playerNumber)
{
	// Load the xml file
	TiXmlDocument *xmlDocCarslist = new TiXmlDocument("cars/carslist.xml");
	if (!xmlDocCarslist->LoadFile())
	{
		delete xmlDocCarslist;
		ScThrowErr("xml file not found");
	}
	
	TiXmlHandle docHandleCarlist(xmlDocCarslist);
	TiXmlElement *xCar = docHandleCarlist.FirstChildElement("carslist").FirstChild("car").Element();
	char *carNames[32];
    int numCars;
    
    for (numCars = 0; xCar != NULL && numCars < 16; numCars++) {
        carNames[numCars] = strdup(xCar->FirstChild()->Value());
        xCar = xCar->NextSiblingElement("car");
    }
    
    char menuTitle[256];
    sprintf(menuTitle, "Choose Player %d Car:", playerNumber);
    
    Menu menu(carNames, numCars, menuTitle);
    int choice = menu.doMenu();
    
	char *returnValue = strdup(carNames[choice]);
	
	for (int i = 0; i < numCars; i++)
		free(carNames[i]);
	
	return returnValue;
}

ToyCar* loadNewToyCar(char *carName, const Tuple& startLocation, double startOrientation)
{
	int i;
	char path[256] = "cars/";
	char filename[256];
    
	strcat(path, carName);
	strcat(path, "/");
	strcpy(filename, path);
	strcat(filename, carName);
	strcat(filename, ".xml");
    
	// Load the xml file
	TiXmlDocument *xmlDocCar = new TiXmlDocument(filename);
	if (!xmlDocCar->LoadFile())
		ScThrowErr("xml file not found");
    
	TiXmlHandle docHandleCar(xmlDocCar);
	TiXmlElement *xSprite = docHandleCar.FirstChildElement("car").FirstChild("sprite").Element();
    int framewidth, frameheight;
    
    if (xSprite == NULL)
        ScThrowErr("xml sprite element not found");
    if (xSprite->QueryIntAttribute("framewidth", &framewidth) != TIXML_SUCCESS)
		ScThrowErr("xml sprite element: frame width attribute not set");    
    if (xSprite->QueryIntAttribute("frameheight", &frameheight) != TIXML_SUCCESS)
		ScThrowErr("xml sprite element: frame height attribute not set");
    
    TiXmlText *xSpriteFilename = docHandleCar.FirstChildElement("car").FirstChild("sprite").FirstChild().Text();
    if (xSpriteFilename == NULL)
        ScThrowErr("xml sprite filename not found");
    
    const int maxConvexes = 16;
	TiXmlElement *xConvex = docHandleCar.FirstChildElement("car").FirstChild("geometry").FirstChild("convex").Element();
    Convex* convexes[maxConvexes];
    short convexCount = 0;
    Tuple points[256];
    short pointsCount;
    while (xConvex != NULL)
    {
        if (convexCount >= maxConvexes)
            ScThrowErr("xml geometry: too many convex objects inside geometry");
        readPointsFromString(xConvex->FirstChild()->Value(), points, pointsCount);
		for (i = 0; i < pointsCount; i++)
			points[i] = points[i] / kScale;
        convexes[convexCount] = new Convex(points, pointsCount);
        xConvex = xConvex->NextSiblingElement("convex");
        convexCount++;
    }
    Geom *geom = new Geom(convexes, convexCount);
    double axelwidth, frontaxel, rearaxel, frontwheelinset, frontwheelradius, rearwheelradius, mass, inertia, frontwheelinertia, rearwheelinertia, drag, rollingresistance;
    
    TiXmlElement *xProperties = docHandleCar.FirstChildElement("car").FirstChild("properties").Element();
    if (xProperties == NULL)
        ScThrowErr("xml properties element not found");
    if (!(  xProperties->Attribute("axelwidth", &axelwidth) &&
            xProperties->Attribute("frontaxel", &frontaxel) &&
            xProperties->Attribute("rearaxel", &rearaxel) &&
            xProperties->Attribute("frontwheelinset", &frontwheelinset) &&
            xProperties->Attribute("frontwheelradius", &frontwheelradius) &&
            xProperties->Attribute("rearwheelradius", &rearwheelradius) &&
            xProperties->Attribute("mass", &mass) &&
            xProperties->Attribute("inertia", &inertia) &&
            xProperties->Attribute("frontwheelinertia", &frontwheelinertia) &&
            xProperties->Attribute("rearwheelinertia", &rearwheelinertia) &&
            xProperties->Attribute("drag", &drag) &&
            xProperties->Attribute("rollingresistance", &rollingresistance)))
    {
		ScThrowErr("xml car properties attributes missing");
    }
    
    int wheelframewidth, wheelframeheight;
    char wheelLayer;
    TiXmlElement *xWheelsprite = docHandleCar.FirstChildElement("car").FirstChild("wheelsprite").Element();
    TiXmlText *xWheelspriteFilename = docHandleCar.FirstChildElement("car").FirstChild("wheelsprite").FirstChild().Text();
    
    if (xWheelsprite == NULL)
        ScThrowErr("xml wheelsprite element not found");
    if (xWheelspriteFilename == NULL)
        ScThrowErr("xml wheelsprite filename not found");
    if (xWheelsprite->QueryIntAttribute("framewidth", &wheelframewidth) != TIXML_SUCCESS)
        ScThrowErr("xml car wheelsprite framewidth missing");
    if (xWheelsprite->QueryIntAttribute("frameheight", &wheelframeheight) != TIXML_SUCCESS)
        ScThrowErr("xml car wheelsprite frameheight missing");
    
    if (xWheelsprite->Attribute("layer") != NULL)
    {
        if (strcmp(xWheelsprite->Attribute("layer"), "above") == 0)
            wheelLayer = 1;
        else if (strcmp(xWheelsprite->Attribute("layer"), "below") == 0)
            wheelLayer = 0;
        else
            ScThrowErr("xml car wheelsprite layer attribute has unknown value");
    }
    else {
        wheelLayer = 0;
    }
    
    TiXmlText *front_SR_Pacejka = docHandleCar.FirstChildElement("car").FirstChild("front_SR_Pacejka").FirstChild().Text();
    TiXmlText *front_SA_Pacejka = docHandleCar.FirstChildElement("car").FirstChild("front_SA_Pacejka").FirstChild().Text();
    TiXmlText *rear_SR_Pacejka = docHandleCar.FirstChildElement("car").FirstChild("rear_SR_Pacejka").FirstChild().Text();
    TiXmlText *rear_SA_Pacejka = docHandleCar.FirstChildElement("car").FirstChild("rear_SA_Pacejka").FirstChild().Text();
    
    if (front_SR_Pacejka == NULL)
        ScThrowErr("xml car front_SR_Pacejka element not found");
    if (front_SA_Pacejka == NULL)
        ScThrowErr("xml car front_SA_Pacejka element not found");
    if (rear_SR_Pacejka == NULL)
        ScThrowErr("xml car rear_SR_Pacejka element not found");
    if (rear_SA_Pacejka == NULL)
        ScThrowErr("xml car rear_SA_Pacejka element not found");
    
    char *word, *brkt, *temp;
    double constants[4];
    
    PacejkaCurve *frontSR, *frontSA, *rearSR, *rearSA;
    
    // front slip ratio Pacejka curve
    temp = strdup(front_SR_Pacejka->Value());
    for (   word = strtok_r(temp, ",", &brkt), i = 0;
            word;
            word = strtok_r(NULL, ",", &brkt), i++)
    {
        if (i >= 4)
            ScThrowErr("xml front SR Pacejka Curve: too many constants.");
        constants[i] = atof(word);
    }
    if (i < 4)
        ScThrowErr("xml front SR Pacejka Curve: too few constants.");
    frontSR = new PacejkaCurve(constants[0], constants[1], constants[2], constants[3]);
    free(temp);
    
    // front slip angle Pacejka curve
    temp = strdup(front_SA_Pacejka->Value());
    for (   word = strtok_r(temp, ",", &brkt), i = 0;
            word;
            word = strtok_r(NULL, ",", &brkt), i++)
    {
        if (i >= 4)
            ScThrowErr("xml front SA Pacejka Curve: too many constants.");
        constants[i] = atof(word);
    }
    if (i < 4)
        ScThrowErr("xml front SA Pacejka Curve: too few constants.");
    frontSA = new PacejkaCurve(constants[0], constants[1], constants[2], constants[3]);
    free(temp);
    
    // rear slip ratio Pacejka curve
    temp = strdup(rear_SR_Pacejka->Value());
    for (   word = strtok_r(temp, ",", &brkt), i = 0;
            word;
            word = strtok_r(NULL, ",", &brkt), i++)
    {
        if (i >= 4)
            ScThrowErr("xml rear SR Pacejka Curve: too many constants.");
        constants[i] = atof(word);
    }
    if (i < 4)
        ScThrowErr("xml rear SR Pacejka Curve: too few constants.");
    rearSR = new PacejkaCurve(constants[0], constants[1], constants[2], constants[3]);
    free(temp);
    
    // rear slip angle Pacejka curve
    temp = strdup(rear_SA_Pacejka->Value());
    for (   word = strtok_r(temp, ",", &brkt), i = 0;
            word;
            word = strtok_r(NULL, ",", &brkt), i++)
    {
        if (i >= 4)
            ScThrowErr("xml rear SA Pacejka Curve: too many constants.");
        constants[i] = atof(word);
    }
    if (i < 4)
        ScThrowErr("xml rear SA Pacejka Curve: too few constants.");
    rearSA = new PacejkaCurve(constants[0], constants[1], constants[2], constants[3]);
    free(temp);
    
    char spritefilename[256];
    char wheelspritefilename[256];
    
	strcpy(spritefilename, path);
    strcat(spritefilename, xSpriteFilename->Value());
    
	strcpy(wheelspritefilename, path);
    strcat(wheelspritefilename, xWheelspriteFilename->Value());
    
    ToyCar* toycar = new ToyCar(    spritefilename,
                                    framewidth,
                                    frameheight,
                                    
                                    wheelspritefilename,
                                    wheelframewidth,
                                    wheelframeheight,
                                    wheelLayer,
                                    
                                    mass,
                                    inertia,
                                    startLocation,
                                    startOrientation,
                                    frontaxel / kScale,
                                    rearaxel / kScale,
                                    axelwidth / kScale,
                                    frontwheelinset / kScale,
                                    frontwheelradius / kScale,
                                    rearwheelradius / kScale,
                                    frontwheelinertia,
                                    rearwheelinertia,
                                    *frontSR,
                                    *frontSA,
                                    *rearSR,
                                    *rearSA,
                                    *geom,
                                    drag,
                                    rollingresistance);
    
    return toycar;
}

Player** players;
/*
void loadPlayers()
{
	// create the players
	players = new Player * [numPlayers];
	ToyCar *toyCar;
	double xpos, ypos, theta;
	int tilewidth = tilemap->getTileWidth();
	int tileheight = tilemap->getTileHeight();
	
	TiXmlElement *xPlayer = docHandle.FirstChildElement("track").FirstChildElement("player").Element();
	for (i = 0; i < numPlayers; i++)
	{
		if (!xPlayer)
		{
			for (j = 0; j < i; j++)
				delete players[j];
			delete [] players;
			xmlDoc->Clear();
			delete xmlDoc;
			ScThrowErr("xml player element not found");
		}
		if (!(xPlayer->Attribute("xpos", &xpos) && xPlayer->Attribute("ypos", &ypos) && xPlayer->Attribute("theta", &theta)))
		{
			for (j = 0; j < i; j++)
				delete players[j];
			delete [] players;
			xmlDoc->Clear();
			delete xmlDoc;
			ScThrowErr("xml player attributes not set");
		}
        
        // check if the sprite is already loaded and load it if not
        toyCar = NULL;
        for (j = 0; j < i; j++) {
            if (strcmp(carNames[j], carNames[i]) == 0) {
                toyCar = new ToyCar(players[j]->car, Tuple(xpos*tilewidth/kScale, ypos*tileheight/kScale), theta);
                break;
            }
        }
        if (toyCar == NULL)
            toyCar = loadNewToyCar(carNames[i], Tuple(xpos*tilewidth/kScale, ypos*tileheight/kScale), theta);
        
		if (i > 0)
			players[i] = new RobotPlayer(*toyCar, pathPolygon, i+1);
		else
			players[i] = new Player(*toyCar, pathPolygon, i+1);
		xPlayer = xPlayer->NextSiblingElement("player");
	}
}
*/
/*
Loads and sets up a game for split-screen up to n players, given the name
of the track folder containing an xml file with the same name describing the track.
*/
CarGame* loadRacingGame(const char* trackfilename)
{
    unsigned int i, j, numPlayers;
    char *items[] = {"New AI Player", "New Human Player", "Done"};
	char menuStr[256];
    Menu menu(items, 3, menuStr);
    int choice;
	list<char*> carNames;
	list<int> playerTypes;
	numPlayers = 0;
	gNumHumanPlayers = 0;
	
	// setup player types and car types
	if (gGameMode == GameMode_HotPotato)
	{
		numPlayers = 2;
		DrawSplash();
		carNames.push_back(chooseToyCar(1));
		DrawSplash();
		carNames.push_back(chooseToyCar(2));
		playerTypes.push_back(1);
		playerTypes.push_back(1);
		gNumHumanPlayers = 2;
	}
	else
	{
		// Display player type selection menus
		while (1)
		{
			DrawSplash();
			sprintf(menuStr, "Add Player %d", numPlayers+1);
			choice = menu.doMenu();
			if (choice == 2) {
				if (numPlayers > 0)
					break;
				else
					continue;
			}
			else if (choice == 1)
				gNumHumanPlayers++;
			numPlayers++;
			DrawSplash();
			carNames.push_back(chooseToyCar(numPlayers));
			playerTypes.push_back(choice);
			if (numPlayers > 15)
				break;
		}
	}
    gNumPlayers = numPlayers;
	
	// draw loading screen
	DrawSplash();
    DrawLoading();
    
	char path[256] = "tracks/";
	char filename[256];
	
	strcat(path, trackfilename);
	strcat(path, "/");
	strcpy(filename, path);
	strcat(filename, trackfilename);
	strcat(filename, ".xml");
	
	// Load the xml file
	TiXmlDocument *xmlDoc = new TiXmlDocument(filename);
	if (!xmlDoc->LoadFile())
	{
		delete xmlDoc;
		ScThrowErr("xml file not found");
	}
	
	TiXmlHandle docHandle(xmlDoc);
	
	TiXmlElement *xTrack = docHandle.FirstChildElement("track").Element();
	TiXmlText *xTileset = docHandle.FirstChildElement("track").FirstChild("tileset").FirstChild().Text();
	TiXmlText *xTilemap = docHandle.FirstChildElement("track").FirstChild("tilemap").FirstChild().Text();
	TiXmlElement *xStartline = docHandle.FirstChildElement("track").FirstChildElement("startline").Element();
	TiXmlText *xPath = docHandle.FirstChildElement("track").FirstChild("path").FirstChild().Text();
	
	// do error checking
	if (!(xTrack && xTileset && xTilemap && xStartline))
	{
		xmlDoc->Clear();
		delete xmlDoc;
		if (!xTrack)
			ScThrowErr("xml track element not found");
		if (!xTileset)
			ScThrowErr("xml tileset element not found");
		if (!xTilemap)
			ScThrowErr("xml tilemap element not found");
		if (!xStartline)
			ScThrowErr("xml startline element not found");
		if (!xPath)
			ScThrowErr("xml path element not found");
	}
	
	// load and create the tileset from an xml file
	ScTilemap* tilemap = loadTileset(xTileset->Value());
	
	// load the tilemap from the specified tilemap file
	strcpy(filename, path);
	strcat(filename, xTilemap->Value());
	tilemap->loadMap(filename);
	
	// the path
	Tuple points[256];
	short pointsCount;
	list<Tuple> *pathPolygon = new list<Tuple>;
    readPointsFromString(xPath->Value(), points, pointsCount);
	for (short k = 0; k < pointsCount; k++) {
		pathPolygon->push_back(points[k] * tilemap->getTileWidth() / kScale);
		//printf("%lf, %lf; ", points[k].x, points[k].y);
	}
	//printf("\n");
	
	// create the players
	players = new Player * [numPlayers];
	ToyCar *toyCar;
	double xpos, ypos, theta;
	int tilewidth = tilemap->getTileWidth();
	int tileheight = tilemap->getTileHeight();
	list<char*>::const_iterator carNamesIterator = carNames.begin();
	list<char*>::const_iterator otherCarNamesIterator;
	list<int>::const_iterator playerTypesIterator = playerTypes.begin();
	otherCarNamesIterator = carNames.begin();
	
	xpos = 0;
	ypos = 0;
	theta = 0;
	for (i = 0; i < numPlayers; i++, carNamesIterator++, playerTypesIterator++)
	{
        // check if the sprite is already loaded and load it if not
        toyCar = NULL;
        for (otherCarNamesIterator = carNames.begin(), j = 0; otherCarNamesIterator != carNamesIterator; otherCarNamesIterator++, j++) {
            if (strcmp(*otherCarNamesIterator, *carNamesIterator) == 0) {
                toyCar = new ToyCar(players[j]->car, Tuple(xpos,ypos), theta);
                break;
            }
        }
		if (toyCar == NULL)
            toyCar = loadNewToyCar(*carNamesIterator, Tuple(xpos,ypos), theta);
		if (*playerTypesIterator == 0)
			players[i] = new RobotPlayer(*toyCar, pathPolygon, i+1);
		else
			players[i] = new Player(*toyCar, pathPolygon, i+1);
	}
	
	// get the start line attributes
	Tuple p1, p2;
	if (!(xStartline->Attribute("x1", &p1.x) && xStartline->Attribute("y1", &p1.y) &&
			xStartline->Attribute("x2", &p2.x) && xStartline->Attribute("y2", &p2.y)))
	{
		for (j = 0; j < numPlayers; j++)
			delete players[j];
		delete [] players;
		xmlDoc->Clear();
		delete xmlDoc;
		ScThrowErr("xml startline attributes not set");
	}
	p1.x *= tilewidth;
	p2.x *= tilewidth;
	p1.y *= tileheight;
	p2.y *= tileheight;
	
	// create the game
	CarGame* game = NULL;
	switch (gGameMode) {
		case GameMode_TimeTrial:
			game = new CarGame(tilemap, players, numPlayers, p1, p2, pathPolygon);
			break;
		case GameMode_HotPotato:
			game = new HotPotatoGame(tilemap, players, numPlayers, p1, p2, gPreferences->getFuseLength(), pathPolygon);
			break;
		case GameMode_KnockEmOut:
			game = new KnockOutGame(tilemap, players, numPlayers, p1, p2, pathPolygon);
			break;
		default:
			printf("error: unknown game mode received\n");
			exit(EXIT_FAILURE);
	}
	if (!game)
	{
		for (j = 0; j < i; j++)
			delete players[j];
		delete [] players;
		xmlDoc->Clear();
		delete xmlDoc;
		ScThrowErr("failed to create the game object");
	}
	
	// Clear and delete our allocated document and return success ;)
	xmlDoc->Clear();
	delete xmlDoc;
	
	printf("successfully created track from %s\n", path);
	
	// add views to the game
	if (gGameMode == GameMode_HotPotato)
	{
		ScPortRect screenRect = {0, 0, gPreferences->getScreenWidth()/2 - 1, gPreferences->getScreenHeight()};
		game->addView(screenRect, 1);
		screenRect.left = gPreferences->getScreenWidth()/2 + 0;
		game->addView(screenRect, 0);
	}
	else//if (gGameMode == GameMode_TimeTrial || gGameMode == GameMode_KnockEmOut)
	{
		ScPortRect screenRect = {0, 0, gPreferences->getScreenWidth(), gPreferences->getScreenHeight()};
		game->addView(screenRect, 0);
	}
	    
	printf("Loading all textures...\n");
	
	// pre-draws all the textures loading them into the vram now
	game->renderEveryTexture();
	
	// clear the screen
	glClear(GL_COLOR_BUFFER_BIT);
	SDL_GL_SwapBuffers();
	glClear(GL_COLOR_BUFFER_BIT);
    
	return game;
}

void InitGL(int width, int height)	        // we call this right after our OpenGL window is created.
{
	// Set up OpenGL for 2D rendering.
	
	glViewport(0, 0, width, height);
	glClearColor(0.0, 0.0, 0.0, 0.0);	// background clear color
	glDisable(GL_DEPTH_TEST);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glColor4ub(255, 255, 255, 255);
    glEnable(GL_POLYGON_SMOOTH);
}

void DrawGLScene()
{
	static double lastTime = GetSeconds();
	static double thisTime;
	
	while ((thisTime = GetSeconds()) - lastTime < 0.001) ;	// slow down on very fast machines
	gGameP->step(std::min(thisTime - lastTime, 0.02));
	lastTime = thisTime;
	
	// swap buffers to display, since we're double buffered.
	SDL_GL_SwapBuffers();
}

void MainLoop(bool autoAccelerate)
{
	bool running;
	SDL_Event event;
	unsigned long startTime, frames;
	
	//printf("autoAccelerate is %d\n", autoAccelerate);
	
	if (autoAccelerate)
	{
		// Start with throttle key in down position
		PlayerEventKey playerEvent;
		playerEvent.type = kPlayerKeyDown;
		playerEvent.key = kThrottleMore;
		for (unsigned int i = 0; i < gNumPlayers; i++)
		{
			players[i]->doEvent(&playerEvent);
		}
	}
	
	// Loop, drawing and checking events
	running = true;
	frames = 0;
	startTime = SDL_GetTicks();
	while (running)
	{
		// get pending events
		while (SDL_PollEvent(&event))
		{
			if (event.type == SDL_QUIT)
				running = false;
			
			if (event.type == SDL_KEYDOWN)
			{
				if (event.key.keysym.sym == SDLK_ESCAPE)
					running = false;
				
				if (event.key.keysym.sym == SDLK_f)
				{
					printf("secs/frame = %f\n", (SDL_GetTicks() - startTime) * 1e-3 / frames);
					startTime = SDL_GetTicks();
					frames = 0;
				}
				
				// pause the game
				if (event.key.keysym.sym == SDLK_p)
				{
					printf("paused... press p to unpause\n");
					DrawPaused();
					while (1) {
						if (SDL_WaitEvent(&event) == 0) {
							printf("error waiting for events\n");
							exit(EXIT_FAILURE);
						}
						if (event.key.keysym.sym == SDLK_p && event.key.type == SDL_KEYUP)
							break;
					}
					while (1) {
						SDL_PollEvent(&event);
						if (event.key.keysym.sym == SDLK_p && event.key.type == SDL_KEYDOWN)
							break;
					}
					continue;
				}
			}
			if (event.type == SDL_KEYDOWN || event.type == SDL_KEYUP)
			{
				PlayerEventKey playerEvent;
				
				switch (event.type) {
					case SDL_KEYDOWN:
						playerEvent.type = kPlayerKeyDown;
						break;
					case SDL_KEYUP:
						playerEvent.type = kPlayerKeyUp;
						switch (event.key.keysym.sym) {
							case SDLK_1:
								if (gNumPlayers < 1)	break;
								printf("Player 1 path distance: %lf.\n", players[0]->getPathDistance());
								printf("Player 1 path location: %lf %lf.\n", players[0]->getPathLocation().x, players[0]->getPathLocation().y);
								break;
							case SDLK_2:
								if (gNumPlayers < 2)	break;
								printf("Player 2 path distance: %lf.\n", players[1]->getPathDistance());
								printf("Player 2 path location: %lf %lf.\n", players[1]->getPathLocation().x, players[1]->getPathLocation().y);
								break;
							case SDLK_3:
								if (gNumPlayers < 3)	break;
								printf("Player 3 path distance: %lf.\n", players[2]->getPathDistance());
								printf("Player 3 path location: %lf %lf.\n", players[2]->getPathLocation().x, players[2]->getPathLocation().y);
								break;
							case SDLK_4:
								if (gNumPlayers < 4)	break;
								printf("Player 4 path distance: %lf.\n", players[3]->getPathDistance());
								printf("Player 4 path location: %lf %lf.\n", players[3]->getPathLocation().x, players[3]->getPathLocation().y);
								break;
							default:
								break;
						}
						break;
				}
				
				switch (event.key.keysym.sym)
				{
					
				// player 0 keys
				
					case SDLK_LEFT:
						playerEvent.key = kSteerWheelLeft;
						gGameP->doEvent(&playerEvent, 0);
						break;
					case SDLK_RIGHT:
						playerEvent.key = kSteerWheelRight;
						gGameP->doEvent(&playerEvent, 0);
						break;
					case SDLK_DOWN:
						playerEvent.key = kThrottleLess;
						gGameP->doEvent(&playerEvent, 0);
						if (autoAccelerate) {
							if (event.type == SDL_KEYUP) {
								playerEvent.type = kPlayerKeyDown;
								playerEvent.key = kThrottleMore;
								gGameP->doEvent(&playerEvent, 0);
							}
							else if (event.type == SDL_KEYDOWN) {
								playerEvent.type = kPlayerKeyUp;
								playerEvent.key = kThrottleMore;
								gGameP->doEvent(&playerEvent, 0);
							}
						}
						break;
					case SDLK_UP:
						if (autoAccelerate)	break;
						playerEvent.key = kThrottleMore;
						gGameP->doEvent(&playerEvent, 0);
						break;
				
				// player 1 keys
				
					case SDLK_a:
						playerEvent.key = kSteerWheelLeft;
						gGameP->doEvent(&playerEvent, 1);
						break;
					case SDLK_d:
						playerEvent.key = kSteerWheelRight;
						gGameP->doEvent(&playerEvent, 1);
						break;
					case SDLK_s:
						playerEvent.key = kThrottleLess;
						gGameP->doEvent(&playerEvent, 1);
						if (autoAccelerate) {
							if (event.type == SDL_KEYUP) {
								playerEvent.type = kPlayerKeyDown;
								playerEvent.key = kThrottleMore;
								gGameP->doEvent(&playerEvent, 1);
							}
							else if (event.type == SDL_KEYDOWN) {
								playerEvent.type = kPlayerKeyUp;
								playerEvent.key = kThrottleMore;
								gGameP->doEvent(&playerEvent, 1);
							}
						}
						break;
					case SDLK_w:
						if (autoAccelerate)	break;
						playerEvent.key = kThrottleMore;
						gGameP->doEvent(&playerEvent, 1);
						break;
				
				// player 2 keys
				
					case SDLK_j:
						playerEvent.key = kSteerWheelLeft;
						gGameP->doEvent(&playerEvent, 2);
						break;
					case SDLK_l:
						playerEvent.key = kSteerWheelRight;
						gGameP->doEvent(&playerEvent, 2);
						break;
					case SDLK_k:
						playerEvent.key = kThrottleLess;
						gGameP->doEvent(&playerEvent, 2);
						if (autoAccelerate) {
							if (event.type == SDL_KEYUP) {
								playerEvent.type = kPlayerKeyDown;
								playerEvent.key = kThrottleMore;
								gGameP->doEvent(&playerEvent, 2);
							}
							else if (event.type == SDL_KEYDOWN) {
								playerEvent.type = kPlayerKeyUp;
								playerEvent.key = kThrottleMore;
								gGameP->doEvent(&playerEvent, 2);
							}
						}
						break;
					case SDLK_i:
						if (autoAccelerate)	break;
						playerEvent.key = kThrottleMore;
						gGameP->doEvent(&playerEvent, 2);
						break;
				
				// player 3 keys
				
					case SDLK_KP4:
						playerEvent.key = kSteerWheelLeft;
						gGameP->doEvent(&playerEvent, 3);
						break;
					case SDLK_KP6:
						playerEvent.key = kSteerWheelRight;
						gGameP->doEvent(&playerEvent, 3);
						break;
					case SDLK_KP5:
						playerEvent.key = kThrottleLess;
						gGameP->doEvent(&playerEvent, 3);
						if (autoAccelerate) {
							if (event.type == SDL_KEYUP) {
								playerEvent.type = kPlayerKeyDown;
								playerEvent.key = kThrottleMore;
								gGameP->doEvent(&playerEvent, 3);
							}
							else if (event.type == SDL_KEYDOWN) {
								playerEvent.type = kPlayerKeyUp;
								playerEvent.key = kThrottleMore;
								gGameP->doEvent(&playerEvent, 3);
							}
						}
						break;
					case SDLK_KP8:
						if (autoAccelerate)	break;
						playerEvent.key = kThrottleMore;
						gGameP->doEvent(&playerEvent, 3);
						break;
					default:
						break;
				}
			}
		}
		
		// Draw
		DrawGLScene();
		frames++;
	}
}
/*
void MainLoop()
{
	bool running;
	SDL_Event event;
    //Uint8 *keys;
	unsigned long startTime, frames;
	
	// Loop, drawing and checking events
	running = true;
	frames = 0;
	startTime = SDL_GetTicks();
	while (running)
	{
		// get pending events
		while (SDL_PollEvent(&event))
		{
			if (event.type == SDL_QUIT)
				running = false;
			/*
			if (event.type == SDL_MOUSEMOTION)
			{
			}
			if (event.type == SDL_MOUSEBUTTONDOWN)
			{
				//printf("clicked at %lf %lf\n", gGameP->views.front().view->getViewRect().left + event.button.x,
				//   gGameP->views.front().view->getViewRect().bottom + gPreferences->getScreenHeight() - event.button.y);
			}
			*
			
			if (event.type == SDL_KEYDOWN)
			{
				if (event.key.keysym.sym == SDLK_ESCAPE)
					running = false;
				
				if (event.key.keysym.sym == SDLK_f)
				{
					printf("secs/frame = %f\n", (SDL_GetTicks() - startTime) * 1e-3 / frames);
					startTime = SDL_GetTicks();
					frames = 0;
				}
				
				// pause the game
				if (event.key.keysym.sym == SDLK_p)
				{
					printf("paused... press p to unpause\n");
					DrawPaused();
					while (1) {
						if (SDL_WaitEvent(&event) == 0) {
							printf("error waiting for events\n");
							exit(EXIT_FAILURE);
						}
						if (event.key.keysym.sym == SDLK_p && event.key.type == SDL_KEYUP)
							break;
					}
					while (1) {
						SDL_PollEvent(&event);
						if (event.key.keysym.sym == SDLK_p && event.key.type == SDL_KEYDOWN)
							break;
					}
					continue;
				}
			}
			if (event.type == SDL_KEYDOWN || event.type == SDL_KEYUP)
			{
				PlayerEventKey playerEvent;
				
				switch (event.type) {
					case SDL_KEYDOWN:
						playerEvent.type = kPlayerKeyDown;
						break;
					case SDL_KEYUP:
						playerEvent.type = kPlayerKeyUp;
						switch (event.key.keysym.sym) {
							case SDLK_1:
								if (gNumPlayers < 1)	break;
								printf("Player 1 path distance: %lf.\n", players[0]->getPathDistance());
								printf("Player 1 path location: %lf %lf.\n", players[0]->getPathLocation().x, players[0]->getPathLocation().y);
								break;
							case SDLK_2:
								if (gNumPlayers < 2)	break;
								printf("Player 2 path distance: %lf.\n", players[1]->getPathDistance());
								printf("Player 2 path location: %lf %lf.\n", players[1]->getPathLocation().x, players[1]->getPathLocation().y);
								break;
							case SDLK_3:
								if (gNumPlayers < 3)	break;
								printf("Player 3 path distance: %lf.\n", players[2]->getPathDistance());
								printf("Player 3 path location: %lf %lf.\n", players[2]->getPathLocation().x, players[2]->getPathLocation().y);
								break;
							case SDLK_4:
								if (gNumPlayers < 4)	break;
								printf("Player 4 path distance: %lf.\n", players[3]->getPathDistance());
								printf("Player 4 path location: %lf %lf.\n", players[3]->getPathLocation().x, players[3]->getPathLocation().y);
								break;
							default:
								break;
						}
						break;
				}
				
				switch (event.key.keysym.sym)
				{
					
				// player 0 keys
				
					case SDLK_LEFT:
						playerEvent.key = kSteerWheelLeft;
						gGameP->doEvent(&playerEvent, 0);
						break;
					case SDLK_RIGHT:
						playerEvent.key = kSteerWheelRight;
						gGameP->doEvent(&playerEvent, 0);
						break;
					case SDLK_UP:
						playerEvent.key = kThrottleMore;
						gGameP->doEvent(&playerEvent, 0);
						break;
					case SDLK_DOWN:
						playerEvent.key = kThrottleLess;
						gGameP->doEvent(&playerEvent, 0);
						break;
				
				// player 1 keys
				
					case SDLK_a:
						playerEvent.key = kSteerWheelLeft;
						gGameP->doEvent(&playerEvent, 1);
						break;
					case SDLK_d:
						playerEvent.key = kSteerWheelRight;
						gGameP->doEvent(&playerEvent, 1);
						break;
					case SDLK_w:
						playerEvent.key = kThrottleMore;
						gGameP->doEvent(&playerEvent, 1);
						break;
					case SDLK_s:
						playerEvent.key = kThrottleLess;
						gGameP->doEvent(&playerEvent, 1);
						break;
				
				// player 2 keys
				
					case SDLK_j:
						playerEvent.key = kSteerWheelLeft;
						gGameP->doEvent(&playerEvent, 2);
						break;
					case SDLK_l:
						playerEvent.key = kSteerWheelRight;
						gGameP->doEvent(&playerEvent, 2);
						break;
					case SDLK_i:
						playerEvent.key = kThrottleMore;
						gGameP->doEvent(&playerEvent, 2);
						break;
					case SDLK_k:
						playerEvent.key = kThrottleLess;
						gGameP->doEvent(&playerEvent, 2);
						break;
				
				// player 3 keys
				
					case SDLK_KP4:
						playerEvent.key = kSteerWheelLeft;
						gGameP->doEvent(&playerEvent, 3);
						break;
					case SDLK_KP6:
						playerEvent.key = kSteerWheelRight;
						gGameP->doEvent(&playerEvent, 3);
						break;
					case SDLK_KP8:
						playerEvent.key = kThrottleMore;
						gGameP->doEvent(&playerEvent, 3);
						break;
					case SDLK_KP5:
						playerEvent.key = kThrottleLess;
						gGameP->doEvent(&playerEvent, 3);
						break;
					default:
						break;
				}
			}
		}
		
		// Draw
		DrawGLScene();
		frames++;
	}
}
*/
void chooseTrack(char result[])
{
	// Load the xml file
	TiXmlDocument *xmlDoc = new TiXmlDocument("tracks/tracklist.xml");
	if (!xmlDoc->LoadFile())
	{
		delete xmlDoc;
		ScThrowErr("xml file not found");
	}
	
	TiXmlHandle docHandle(xmlDoc);
	TiXmlElement *xTrack = docHandle.FirstChildElement("tracklist").FirstChild("track").Element();
	char *trackNames[16];
    int numTracks;
    
    for (numTracks = 0; xTrack != NULL && numTracks < 16; numTracks++) {
        trackNames[numTracks] = strdup(xTrack->Attribute("name"));
        xTrack = xTrack->NextSiblingElement("track");
    }
    
    Menu menu(trackNames, numTracks, "Choose track:");
    int choice = menu.doMenu();
    strcpy(result, trackNames[choice]);
    for (int i = 0; i < numTracks; i++)
        free(trackNames[i]);
}

void musicDone() {
	Mix_HaltMusic();
	Mix_FreeMusic(gMusic);
	gMusic = NULL;
}

int main(int argc, char **argv) 
{
	unsigned long flags;
	
	// Set flags
	flags = SDL_OPENGL;
	
	// Handle command line options
	for (int i = 1; argc > 1; i++, argc--)
	{
		if(strncmp(argv[i], "-datadir", 8) == 0)
		{
			if (argv[i][8] != '=')
			{
				fprintf(stderr, "error: missing argument to -datadir\n");
				exit(EXIT_FAILURE);
			}
			if (strlen(&argv[i][9]) > 255)
			{
				fprintf(stderr, "error: directory path cannot be more than 255 characters long for -datadir\n");
				exit(EXIT_FAILURE);
			}
			strncpy(gDataDir, &argv[i][9], 255);
		}
		else if(strcmp(argv[i], "-h") == 0)
		{
			printf("Usage: %s [OPTION]\n\n", argv[0]);
			printf("Options:\n");
			printf("   -datadir=DIR   Specify the location of the game data directory\n");
			printf("   -h             Display this help text\n");
			exit(0);
		}
		else
		{
			fprintf(stderr, "warning: unrecognised argument \"%s\"\n", argv[i]);
		}
	}
	
    // Load preferences
    try {
    char str[256];
    
    // Check for directory. Create if it does not exist.
    int cwd = open(".", O_RDONLY);
    sprintf(str, "%s/.toycars", getenv("HOME"));
    if (chdir(str) == -1) {
        if (errno == ENOENT) {
            // directory does not exist so create it
            sprintf(str, "mkdir %s/.toycars", getenv("HOME"));
            system(str);
        }
        else {
            // an error of some kind occured
            ScThrowErr("chdir command encountered an error");
        }
    }
    fchdir(cwd);
    close(cwd);
    
    sprintf(str, "%s/.toycars/preferences.xml", getenv("HOME"));
    gPreferences = new TcPreferences(str);
	
	if (gPreferences->getFullscreen())
		flags |= SDL_FULLSCREEN;
    }
    catch (ScException e)
    {
        e.printMsg();
        exit(EXIT_FAILURE);
    }
    
	// Initialize SDL for video output
	if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_AUDIO) < 0)
	{
		fprintf(stderr, "Unable to initialize SDL: %s\n", SDL_GetError());
		exit(EXIT_FAILURE);
	}
	
	// Create an OpenGL screen
	if (SDL_SetVideoMode(gPreferences->getScreenWidth(), gPreferences->getScreenHeight(), 0, flags) == NULL)
	{
		fprintf(stderr, "Unable to create OpenGL screen: %s\n", SDL_GetError());
		SDL_Quit();
		exit(EXIT_FAILURE);
	}
	
	// Set the title bar in environments that support it
	SDL_WM_SetCaption("toycars", NULL);
    
	// Output instructions
	printf("toycars\n");
	printf("Arrows: steering and throttle for player 1\n");
	printf("adws: steering and throttle for player 2\n");
	//printf("Page up and down: zoom view\n");
	//printf("Spacebar: shoot\n");
	printf("f: aproximate seconds per frame\n");
	printf("p: pause\n");
	printf("Esc: quit\n");
	
	// Seed the random number generator
	srand(time(NULL));
	
	// Initialize the OpenAL library
	// Do this before creating the CarGame object
	//alutInit(&argc, argv);
    
	// Initialise OpenGL
	//printf("Changing to directory: %s\n", gDataDir);
	if (chdir(gDataDir) == -1) {
		printf("chdir to directory \"%s\" failed.\n", gDataDir);
		exit(EXIT_FAILURE);
	}
	if (chdir("data") == -1) {
		printf("chdir to directory \"data\" failed.\n");
		exit(EXIT_FAILURE);
	}
	
	/* Same setup as before */
	int audio_rate = MIX_DEFAULT_FREQUENCY;
	Uint16 audio_format = AUDIO_S16; 
	int audio_channels = 2;
	int audio_buffers = 4096;
	
	if(Mix_OpenAudio(audio_rate, audio_format, audio_channels, audio_buffers)) {
		printf("Unable to open audio!\n");
		exit(1);
	}
		
	//bool running = 1;
	//pthread_t musicThread;
	//pthread_create(&musicThread, NULL, streamMusic, &running);
	
	// initialise the glf library
	// must be done before calling DrawSplash
	glfInit();
	int fontID = glfLoadFont("fonts/cricket1.glf");
	if (fontID == GLF_ERROR) {
		printf("glfLoadFont returned GLF_ERROR\n");
        exit(EXIT_FAILURE);
    }
	glfSetCurrentFont(fontID);
    
    int menuResult;
    char trackName[256];
    char *items[] = {"Timed Race", "Hot Potato (2P)", "Knock 'em Out", "Preferences", "Credits", "Quit"};
    Menu menu(items, 6);
	bool quitting = false;
	
    do {
        InitGL(gPreferences->getScreenWidth(), gPreferences->getScreenHeight());
        DrawSplash();
		//SDL_GL_SwapBuffers();
		if (Mix_PlayingMusic() == 0) {
			gMusic = Mix_LoadMUS(kTitleSongFileName);
			if (gMusic == NULL)
				printf("failed to load music\n");
			Mix_VolumeMusic(MIX_MAX_VOLUME/2);
			Mix_HookMusicFinished(musicDone);
			Mix_PlayMusic(gMusic, 0);
		}
        menuResult = menu.doMenu();
        //printf("menu result is: %d\n", menuResult);
        switch (menuResult)
        {
            case 0:     // Time Trial
				gGameMode = GameMode_TimeTrial;
                break;
			case 1:		// Hot Potato
				gGameMode = GameMode_HotPotato;
				break;
            case 2:     // Knock Em Out
				gGameMode = GameMode_KnockEmOut;
                break;
            case 3:     // Preferences
				gPreferences->drawPrefs();
				if (gPreferences->getModifiedScreenPrefs()) {
					if (SDL_SetVideoMode(gPreferences->getScreenWidth(), gPreferences->getScreenHeight(), 0, SDL_OPENGL | (gPreferences->getFullscreen() ? SDL_FULLSCREEN : 0)) == NULL)
					{
						fprintf(stderr, "Unable to create OpenGL screen: %s\n", SDL_GetError());
						SDL_Quit();
						exit(EXIT_FAILURE);
					}
				}
				continue;
            case 4:     // Credits
				DrawSplash();
				ShowCredits();
				continue;
            case 5:     // Quit
				quitting = true;
                continue;
			default:
				printf("error: unknown menu item");
				exit(EXIT_FAILURE);
        }
        
        DrawSplash();
        chooseTrack(trackName);
        
        printf("Setting up animation...\n");
        
        // load tiles tilemap and car sprites from track xml file
       	try { gGameP = loadRacingGame(trackName); }
        catch (ScException e)
        {
            e.printMsg();
            if (gGameP)	delete gGameP;
            SDL_Quit();
            exit(EXIT_FAILURE);
        }
        
        // Enter the main loop
		MainLoop(gNumHumanPlayers > 2);
		/*
		//Mix_FadeOutMusic(2000);
        if (gNumPlayers <= 2)
			MainLoop();
        else //if (gNumPlayers == 4)
			MainLoopFourPlayers();
		//else
		//	printf("error: number of players: %d not supported in this version", gNumPlayers);
		*/
        delete gGameP;
    } while (quitting == false);
    
	printf("Quitting...\n");
	musicDone();
	//running = 0;
	//pthread_join(musicThread, NULL);
	
	// Clean up
	glfUnloadFont();
    delete gPreferences;
	
	//alutExit();
	Mix_CloseAudio();
	SDL_Quit();
	
	return 0;
}
/*
void *streamMusic(void *arg)
{
	bool *running = (bool *)arg;
	ogg_stream musicStreamer;
	
	try
	{
		printf("Loading title music...\n");
		musicStreamer.open(kTitleSongFileName);
		musicStreamer.display();
		
		while (musicStreamer.update() && *running)
		{
			if(!musicStreamer.playing())
			{
				if(!musicStreamer.playback())
					throw string("Ogg abruptly stopped.");
				else
					cout << "Ogg stream was interrupted.\n";
			}
		}
	}
	catch(string error)
	{
		cout << error;
		cin.get();
	}
	
	musicStreamer.release();
	
	return NULL;
}
*/
